@@ -1,12 +1,13 @@
+from __future__ import annotations
+
 import dataclasses
 import os
-from typing import ClassVar, Optional
+from typing import TYPE_CHECKING, ClassVar
 
 from loguru import logger
 
-from daft.runners.pyrunner import PyRunner
-from daft.runners.ray_runner import RayRunner
-from daft.runners.runner import Runner
+if TYPE_CHECKING:
+    from daft.runners.runner import Runner
 
 
 class _RunnerConfig:
@@ -21,7 +22,7 @@ class _PyRunnerConfig(_RunnerConfig):
 @dataclasses.dataclass(frozen=True)
 class _RayRunnerConfig(_RunnerConfig):
     name = "ray"
-    address: Optional[str]
+    address: str | None
 
 
 def _get_runner_config_from_env() -> _RunnerConfig:
@@ -43,7 +44,7 @@ def _get_runner_config_from_env() -> _RunnerConfig:
 
 
 # Global Runner singleton, initialized when accessed through the DaftContext
-_RUNNER: Optional[Runner] = None
+_RUNNER: Runner | None = None
 
 
 @dataclasses.dataclass(frozen=True)
@@ -58,10 +59,14 @@ class DaftContext:
         if _RUNNER is not None:
             return _RUNNER
         if self.runner_config.name == "ray":
+            from daft.runners.ray_runner import RayRunner
+
             logger.info("Using RayRunner")
             assert isinstance(self.runner_config, _RayRunnerConfig)
             _RUNNER = RayRunner(address=self.runner_config.address)
         elif self.runner_config.name == "py":
+            from daft.runners.pyrunner import PyRunner
+
             logger.info("Using PyRunner")
             _RUNNER = PyRunner()
         else:
@@ -85,7 +90,7 @@ def get_context() -> DaftContext:
     return _DaftContext
 
 
-def set_runner_ray(address: Optional[str] = None) -> DaftContext:
+def set_runner_ray(address: str | None = None) -> DaftContext:
     """Sets the runner for executing Daft dataframes to the RayRunner
 
     Alternatively, users can set this behavior via environment variables:
