@@ -1,21 +1,9 @@
 from __future__ import annotations
 
 import io
-import uuid
 from dataclasses import dataclass
 from functools import partial
-from typing import (
-    IO,
-    Any,
-    Callable,
-    Dict,
-    Iterable,
-    List,
-    Optional,
-    Tuple,
-    TypeVar,
-    Union,
-)
+from typing import IO, TYPE_CHECKING, Any, Callable, Iterable, TypeVar, Union
 
 import pandas
 import pyarrow as pa
@@ -26,19 +14,25 @@ from daft.context import get_context
 from daft.dataframe.schema import DataFrameSchema
 from daft.datasources import (
     CSVSourceInfo,
-    InMemorySourceInfo,
     JSONSourceInfo,
     ParquetSourceInfo,
     StorageType,
 )
+from daft.errors import ExpressionTypeError
 from daft.execution.operators import ExpressionType
 from daft.expressions import ColumnExpression, Expression, col
-from daft.filesystem import get_filesystem_from_path
+from daft.filesystem import get_filesystem_from_path, get_protocol_from_path
 from daft.logical import logical_plan
 from daft.logical.schema import ExpressionList
-from daft.runners.partitioning import PartitionSet
+from daft.runners.partitioning import PartitionCacheEntry, PartitionSet, vPartition
+from daft.runners.pyrunner import LocalPartitionSet
+from daft.runners.ray_runner import RayPartitionSet
+from daft.types import PythonExpressionType
 from daft.viz import DataFrameDisplay
 
+if TYPE_CHECKING:
+    from ray.data.dataset import Dataset as RayDataset
+
 UDFReturnType = TypeVar("UDFReturnType", covariant=True)
 
 ColumnInputType = Union[Expression, str]
@@ -67,11 +61,12 @@ def _sample_with_pyarrow(
 
 def _get_filepaths(path: str):
     fs = get_filesystem_from_path(path)
+    protocol = get_protocol_from_path(path)
     if fs.isdir(path):
-        return fs.ls(path)
+        return [f"{protocol}://{path}" if protocol != "file" else path for path in fs.ls(path)]
     elif fs.isfile(path):
         return [path]
-    return fs.expand_path(path, recursive=True)
+    return [f"{protocol}://{path}" if protocol != "file" else path for path in fs.expand_path(path, recursive=True)]
 
 
 class DataFrame:
@@ -86,8 +81,24 @@ class DataFrame:
         Args:
             plan: LogicalPlan describing the steps required to arrive at this DataFrame
         """
-        self._plan = plan
-        self._result: Optional[PartitionSet] = None
+        self.__plan = plan
+        self._result_cache: PartitionCacheEntry | None = None
+
+    @property
+    def _plan(self) -> logical_plan.LogicalPlan:
+        if self._result_cache is None:
+            return self.__plan
+        else:
+            return logical_plan.InMemoryScan(
+                self._result_cache, self.__plan.schema().to_column_expressions(), self.__plan.partition_spec()
+            )
+
+    @property
+    def _result(self) -> PartitionSet | None:
+        if self._result_cache is None:
+            return None
+        else:
+            return self._result_cache.value
 
     def plan(self) -> logical_plan.LogicalPlan:
         """Returns `LogicalPlan` that will be executed to compute the result of this DataFrame.
@@ -95,10 +106,29 @@ class DataFrame:
         Returns:
             logical_plan.LogicalPlan: LogicalPlan to compute this DataFrame.
         """
-        return self._plan
+        return self.__plan
+
+    def explain(self, show_optimized: bool = False) -> None:
+        """Prints the LogicalPlan that will be executed to produce this DataFrame.
+        Defaults to showing the unoptimized plan. Use `show_optimized` to show the optimized one.
+
+        Args:
+            show_optimized (bool): shows the optimized QueryPlan instead of the unoptimized one.
+        """
+
+        if self._result_cache is not None:
+            print("Result is cached and will skip computation\n")
+            print(self._plan.pretty_print())
+
+            print("However here is the logical plan used to produce this result:\n")
+
+        plan = self.__plan
+        if show_optimized:
+            plan = get_context().runner().optimize(plan)
+        print(plan.pretty_print())
 
     def num_partitions(self) -> int:
-        return self._plan.num_partitions()
+        return self.__plan.num_partitions()
 
     def schema(self) -> DataFrameSchema:
         """Returns the DataFrameSchema of the DataFrame, which provides information about each column
@@ -106,17 +136,27 @@ class DataFrame:
         Returns:
             DataFrameSchema: schema of the DataFrame
         """
-        return DataFrameSchema.from_expression_list(self._plan.schema())
+        return DataFrameSchema.from_expression_list(self.__plan.schema())
 
-    def column_names(self) -> List[str]:
+    @property
+    def column_names(self) -> list[str]:
         """Returns column names of DataFrame as a list of strings.
 
         Returns:
             List[str]: Column names of this DataFrame.
         """
-        return [expr.name() for expr in self._plan.schema()]
+        return [expr.name() for expr in self.__plan.schema()]
 
-    def show(self, n: int = -1) -> DataFrameDisplay:
+    @property
+    def columns(self) -> list[ColumnExpression]:
+        """Returns column of DataFrame as a list of ColumnExpressions.
+
+        Returns:
+            List[ColumnExpression]: Columns of this DataFrame.
+        """
+        return [expr.to_column_expression() for expr in self.__plan.schema()]
+
+    def show(self, n: int | None = None) -> DataFrameDisplay:
         """Executes and displays the executed dataframe as a table
 
         Args:
@@ -129,8 +169,9 @@ class DataFrame:
                 This call is **blocking** and will execute the DataFrame when called
         """
         df = self
-        if n != -1:
+        if n is not None:
             df = df.limit(n)
+
         execution_result = df.to_pandas()
         return DataFrameDisplay(execution_result, df.schema())
 
@@ -145,7 +186,7 @@ class DataFrame:
     ###
 
     @classmethod
-    def from_pylist(cls, data: List[Dict[str, Any]]) -> DataFrame:
+    def from_pylist(cls, data: list[dict[str, Any]]) -> DataFrame:
         """Creates a DataFrame from a list of dictionaries
 
         Example:
@@ -159,41 +200,62 @@ class DataFrame:
         """
         if not data:
             raise ValueError("Unable to create DataFrame from empty list")
-        schema = ExpressionList(
-            [
-                ColumnExpression(header, expr_type=ExpressionType.from_py_type(type(data[0][header])))
-                for header in data[0]
-            ]
-        )
-        plan = logical_plan.Scan(
-            schema=schema,
-            predicate=None,
-            columns=None,
-            source_info=InMemorySourceInfo(data={header: [row[header] for row in data] for header in data[0]}),
-        )
-        return cls(plan)
+        return cls.from_pydict(data={header: [row[header] for row in data] for header in data[0]})
 
     @classmethod
-    def from_pydict(cls, data: Dict[str, Any]) -> DataFrame:
-        """Creates a DataFrame from an In-Memory Data columnar source that is passed in as `data`.
+    def from_pydict(cls, data: dict[str, list | pa.Array]) -> DataFrame:
+        """Creates a DataFrame from a Python Dictionary.
 
         Example:
             >>> df = DataFrame.from_pydict({"foo": [1, 2]})
 
         Args:
-            data: Key -> Sequence[item] of data. Each Key is created as a column.
+            data: Key -> Sequence[item] of data. Each Key is created as a column, and must have a value that is
+                either a Python list or PyArrow array. Values must be equal in length across all keys.
 
         Returns:
             DataFrame: DataFrame created from dictionary of columns
         """
+
+        block_data: dict[str, tuple[ExpressionType, Any]] = {}
+        for header in data:
+            arr = data[header]
+
+            if isinstance(arr, pa.Array) or isinstance(arr, pa.ChunkedArray):
+                expr_type = ExpressionType.from_arrow_type(arr.type)
+                block_data[header] = (expr_type, arr.to_pylist() if ExpressionType.is_py(expr_type) else arr)
+                continue
+
+            try:
+                arrow_type = pa.infer_type(arr)
+            except pa.lib.ArrowInvalid:
+                arrow_type = None
+
+            if arrow_type is None or pa.types.is_nested(arrow_type):
+                found_types = {type(o) for o in data[header]} - {type(None)}
+                block_data[header] = (
+                    (ExpressionType.python_object(), list(arr))
+                    if len(found_types) > 1
+                    else (PythonExpressionType(found_types.pop()), list(arr))
+                )
+                continue
+
+            expr_type = ExpressionType.from_arrow_type(arrow_type)
+            block_data[header] = (expr_type, list(arr) if ExpressionType.is_py(expr_type) else pa.array(arr))
+
         schema = ExpressionList(
-            [ColumnExpression(header, expr_type=ExpressionType.from_py_type(type(data[header][0]))) for header in data]
+            [ColumnExpression(header, expr_type=expr_type) for header, (expr_type, _) in block_data.items()]
+        ).resolve()
+        data_vpartition = vPartition.from_pydict(
+            data={header: arr for header, (_, arr) in block_data.items()}, schema=schema, partition_id=0
         )
-        plan = logical_plan.Scan(
+        result_pset = LocalPartitionSet({0: data_vpartition})
+
+        cache_entry = get_context().runner().put_partition_set_into_cache(result_pset)
+
+        plan = logical_plan.InMemoryScan(
+            cache_entry=cache_entry,
             schema=schema,
-            predicate=None,
-            columns=None,
-            source_info=InMemorySourceInfo(data=data),
         )
         return cls(plan)
 
@@ -236,7 +298,7 @@ class DataFrame:
         cls,
         path: str,
         has_headers: bool = True,
-        column_names: Optional[List[str]] = None,
+        column_names: list[str] | None = None,
         delimiter: str = ",",
     ) -> DataFrame:
         """Creates a DataFrame from CSV file(s)
@@ -335,6 +397,9 @@ class DataFrame:
     def from_files(cls, path: str) -> DataFrame:
         """Creates a DataFrame from files in storage, where each file is one row of the DataFrame
 
+        Example:
+            >>> df = DataFrame.from_files("/path/to/files/*.jpeg")
+
         Args:
             path (str): path to files on disk (allows wildcards)
 
@@ -351,7 +416,7 @@ class DataFrame:
     ###
 
     def write_parquet(
-        self, root_dir: str, compression: str = "snappy", partition_cols: Optional[List[ColumnInputType]] = None
+        self, root_dir: str, compression: str = "snappy", partition_cols: list[ColumnInputType] | None = None
     ) -> DataFrame:
         """Writes the DataFrame to parquet files using a `root_dir` and randomly generated UUIDs as the filepath and returns the filepaths.
         Currently generates a parquet file per partition unless `partition_cols` are used, then the number of files can equal the number of partitions times the number of values of partition col.
@@ -367,7 +432,7 @@ class DataFrame:
             .. NOTE::
                 This call is **blocking** and will execute the DataFrame when called
         """
-        cols: Optional[ExpressionList] = None
+        cols: ExpressionList | None = None
         if partition_cols is not None:
             cols = self.__column_input_to_expression(tuple(partition_cols))
             for c in cols:
@@ -387,10 +452,9 @@ class DataFrame:
         write_df = DataFrame(plan)
         write_df.collect()
         assert write_df._result is not None
-        data = write_df._result.to_pydict()
-        return DataFrame.from_pydict(data)
+        return DataFrame(write_df._plan)
 
-    def write_csv(self, root_dir: str, partition_cols: Optional[List[ColumnInputType]] = None) -> DataFrame:
+    def write_csv(self, root_dir: str, partition_cols: list[ColumnInputType] | None = None) -> DataFrame:
         """Writes the DataFrame to CSV files using a `root_dir` and randomly generated UUIDs as the filepath and returns the filepaths.
         Currently generates a csv file per partition unless `partition_cols` are used, then the number of files can equal the number of partitions times the number of values of partition col.
 
@@ -405,7 +469,7 @@ class DataFrame:
             .. NOTE::
                 This call is **blocking** and will execute the DataFrame when called
         """
-        cols: Optional[ExpressionList] = None
+        cols: ExpressionList | None = None
         if partition_cols is not None:
             cols = self.__column_input_to_expression(tuple(partition_cols))
             for c in cols:
@@ -424,8 +488,7 @@ class DataFrame:
         write_df = DataFrame(plan)
         write_df.collect()
         assert write_df._result is not None
-        data = write_df._result.to_pydict()
-        return DataFrame.from_pydict(data)
+        return DataFrame(write_df._plan)
 
     ###
     # DataFrame operations
@@ -435,6 +498,47 @@ class DataFrame:
         expressions = [col(c) if isinstance(c, str) else c for c in columns]
         return ExpressionList(expressions)
 
+    def __getitem__(self, item: slice | int | str | Iterable[str | int]) -> ColumnExpression | DataFrame:
+        result: ColumnExpression | None
+
+        if isinstance(item, int):
+            exprs = self._plan.schema()
+            if item < -len(exprs.exprs) or item >= len(exprs.exprs):
+                raise ValueError(f"{item} out of bounds for {exprs.exprs}")
+            result = exprs.exprs[item]
+            assert result is not None
+            return result.to_column_expression()
+        elif isinstance(item, str):
+            exprs = self._plan.schema()
+            result = exprs.get_expression_by_name(item)
+            if result is None:
+                raise ValueError(f"{item} not found in DataFrame schema {exprs}")
+            assert result is not None
+            return result.to_column_expression()  # type: ignore
+        elif isinstance(item, Iterable):
+            exprs = self._plan.schema()
+            columns = []
+            for it in item:
+                if isinstance(it, str):
+                    result = exprs.get_expression_by_name(it)
+                    if result is None:
+                        raise ValueError(f"{it} not found in DataFrame schema {exprs}")
+                    columns.append(result)
+                elif isinstance(it, int):
+                    if it < -len(exprs.exprs) or it >= len(exprs.exprs):
+                        raise ValueError(f"{it} out of bounds for {exprs.exprs}")
+                    result = exprs.exprs[it]
+                    assert result is not None
+                    columns.append(result.to_column_expression())
+                else:
+                    raise ValueError(f"unknown indexing type: {type(it)}")
+            return self.select(*columns)
+        elif isinstance(item, slice):
+            exprs = self._plan.schema()
+            return self.select(*[val.to_column_expression() for val in exprs.exprs[item]])
+        else:
+            raise ValueError(f"unknown indexing type: {type(item)}")
+
     def select(self, *columns: ColumnInputType) -> DataFrame:
         """Creates a new DataFrame that `selects` that columns that are passed in from the current DataFrame.
 
@@ -472,10 +576,16 @@ class DataFrame:
             DataFrame: DataFrame that has only  unique rows.
         """
         all_exprs = self._plan.schema()
-        gb = self.groupby(*[col(e.name()) for e in all_exprs])
-        first_e_name = [e.name() for e in all_exprs][0]
-        dummy_col_name = str(uuid.uuid4())
-        return gb.agg([(col(first_e_name).alias(dummy_col_name), "min")]).exclude(dummy_col_name)
+        plan: logical_plan.LogicalPlan = logical_plan.LocalDistinct(self._plan, all_exprs)
+        if self.num_partitions() > 1:
+            plan = logical_plan.Repartition(
+                plan,
+                partition_by=all_exprs,
+                num_partitions=self.num_partitions(),
+                scheme=logical_plan.PartitionScheme.HASH,
+            )
+            plan = logical_plan.LocalDistinct(plan, all_exprs)
+        return DataFrame(plan)
 
     def exclude(self, *names: str) -> DataFrame:
         """Drops columns from the current DataFrame by name.
@@ -530,9 +640,7 @@ class DataFrame:
         )
         return DataFrame(projection)
 
-    def sort(
-        self, by: Union[ColumnInputType, List[ColumnInputType]], desc: Union[bool, List[bool]] = False
-    ) -> DataFrame:
+    def sort(self, by: ColumnInputType | list[ColumnInputType], desc: bool | list[bool] = False) -> DataFrame:
         """Sorts DataFrame globally according to column.
 
         Example:
@@ -605,9 +713,9 @@ class DataFrame:
     def join(
         self,
         other: DataFrame,
-        on: Optional[Union[List[ColumnInputType], ColumnInputType]] = None,
-        left_on: Optional[Union[List[ColumnInputType], ColumnInputType]] = None,
-        right_on: Optional[Union[List[ColumnInputType], ColumnInputType]] = None,
+        on: list[ColumnInputType] | ColumnInputType | None = None,
+        left_on: list[ColumnInputType] | ColumnInputType | None = None,
+        right_on: list[ColumnInputType] | ColumnInputType | None = None,
         how: str = "inner",
     ) -> DataFrame:
         """Joins left (self) DataFrame on the right on a set of keys.
@@ -691,11 +799,11 @@ class DataFrame:
         exprs_to_explode = self.__column_input_to_expression(columns)
         explode_op = logical_plan.Explode(
             self._plan,
-            exprs_to_explode,
+            ExpressionList([e._explode() for e in exprs_to_explode]),
         )
         return DataFrame(explode_op)
 
-    def _agg(self, to_agg: List[Tuple[ColumnInputType, str]], group_by: Optional[ExpressionList] = None) -> DataFrame:
+    def _agg(self, to_agg: list[tuple[ColumnInputType, str]], group_by: ExpressionList | None = None) -> DataFrame:
         assert len(to_agg) > 0, "no columns to aggregate."
         exprs_to_agg = self.__column_input_to_expression(tuple(e for e, _ in to_agg))
         ops = [op for _, op in to_agg]
@@ -703,10 +811,20 @@ class DataFrame:
         function_lookup = {
             "sum": Expression._sum,
             "count": Expression._count,
+            "mean": Expression._mean,
             "min": Expression._min,
             "max": Expression._max,
-            "count": Expression._count,
         }
+
+        if self.num_partitions() == 1:
+            agg_exprs = []
+
+            for e, op_name in zip(exprs_to_agg, ops):
+                assert op_name in function_lookup
+                agg_exprs.append((function_lookup[op_name](e).alias(e.name()), op_name))
+            plan = logical_plan.LocalAggregate(self._plan, agg=agg_exprs, group_by=group_by)
+            return DataFrame(plan)
+
         intermediate_ops = {
             "sum": ("sum",),
             "count": ("count",),
@@ -719,9 +837,9 @@ class DataFrame:
 
         finalizer_ops_funcs = {"mean": lambda x, y: (x + 0.0) / (y + 0.0)}
 
-        first_phase_ops: List[Tuple[Expression, str]] = []
-        second_phase_ops: List[Tuple[Expression, str]] = []
-        finalizer_phase_ops: List[Expression] = []
+        first_phase_ops: list[tuple[Expression, str]] = []
+        second_phase_ops: list[tuple[Expression, str]] = []
+        finalizer_phase_ops: list[Expression] = []
         need_final_projection = False
         for e, op in zip(exprs_to_agg, ops):
             assert op in intermediate_ops
@@ -818,6 +936,60 @@ class DataFrame:
         assert len(cols) > 0, "no columns were passed in"
         return self._agg([(c, "mean") for c in cols])
 
+    def min(self, *cols: ColumnInputType) -> DataFrame:
+        """Performs a global min on the DataFrame on a sequence of columns.
+
+        Args:
+            *cols (Union[str, Expression]): columns to min
+        Returns:
+            DataFrame: Globally aggregated min. Should be a single row.
+        """
+        assert len(cols) > 0, "no columns were passed in"
+        return self._agg([(c, "min") for c in cols])
+
+    def max(self, *cols: ColumnInputType) -> DataFrame:
+        """Performs a global max on the DataFrame on a sequence of columns.
+
+        Args:
+            *cols (Union[str, Expression]): columns to max
+        Returns:
+            DataFrame: Globally aggregated max. Should be a single row.
+        """
+        assert len(cols) > 0, "no columns were passed in"
+        return self._agg([(c, "max") for c in cols])
+
+    def count(self, *cols: ColumnInputType) -> DataFrame:
+        """Performs a global count on the DataFrame on a sequence of columns.
+
+        Args:
+            *cols (Union[str, Expression]): columns to count
+        Returns:
+            DataFrame: Globally aggregated count. Should be a single row.
+        """
+        assert len(cols) > 0, "no columns were passed in"
+        return self._agg([(c, "count") for c in cols])
+
+    def agg(self, to_agg: list[tuple[ColumnInputType, str]]) -> DataFrame:
+        """Perform aggregations on this DataFrame. Allows for mixed aggregations for multiple columns
+        Will return a single row that aggregated the entire DataFrame.
+
+        Example:
+        >>> df = df.agg([
+        >>>     ('x', 'sum'),
+        >>>     ('x', 'mean'),
+        >>>     ('y', 'min'),
+        >>>     ('y', 'max'),
+        >>>     (col('x') + col('y'), 'max'),
+        >>> ])
+
+        Args:
+            to_agg (List[Tuple[ColumnInputType, str]]): list of (column, agg_type)
+
+        Returns:
+            DataFrame: DataFrame with aggregated results
+        """
+        return self._agg(to_agg, group_by=None)
+
     def groupby(self, *group_by: ColumnInputType) -> GroupedDataFrame:
         """Performs a GroupBy on the DataFrame for Aggregation.
 
@@ -837,7 +1009,10 @@ class DataFrame:
         """
         context = get_context()
         if self._result is None:
-            self._result = context.runner().run(self._plan)
+            self._result_cache = context.runner().run(self._plan)
+            result = self._result
+            assert result is not None
+            result.wait()
         return self
 
     def to_pandas(self) -> pandas.DataFrame:
@@ -851,18 +1026,54 @@ class DataFrame:
                 This call is **blocking** and will execute the DataFrame when called
         """
         self.collect()
-        assert self._result is not None
-        pd_df = self._result.to_pandas(schema=self._plan.schema())
-        del self._result
-        self._result = None
+        result = self._result
+        assert result is not None
+
+        pd_df = result.to_pandas(schema=self._plan.schema())
         return pd_df
 
+    def to_pydict(self) -> pandas.DataFrame:
+        """Converts the current DataFrame to a python dict.
+        If results have not computed yet, collect will be called.
+
+        Returns:
+            dict[str, list[Any]]: python dict converted from a Daft DataFrame
+
+            .. NOTE::
+                This call is **blocking** and will execute the DataFrame when called
+        """
+        self.collect()
+        result = self._result
+        assert result is not None
+        return result.to_pydict()
+
+    def to_ray_dataset(self) -> RayDataset:
+        """Converts the current DataFrame to a Ray Dataset which is useful for running distributed ML model training in Ray
+
+        .. NOTE::
+            This function can only work if Daft is running using the RayRunner
+
+        Returns:
+            RayDataset: Ray dataset
+        """
+        self.collect()
+        partition_set = self._result
+        assert partition_set is not None
+        assert isinstance(partition_set, RayPartitionSet), "Cannot convert to Ray Dataset if not running on Ray backend"
+        return partition_set.to_ray_dataset()
+
 
 @dataclass
 class GroupedDataFrame:
     df: DataFrame
     group_by: ExpressionList
 
+    def __post_init__(self):
+        resolved_groupby = self.group_by.resolve(self.df._plan.schema())
+        for e in resolved_groupby:
+            if e.resolved_type() == ExpressionType.null():
+                raise ExpressionTypeError(f"Cannot groupby on null type expression: {e}")
+
     def sum(self, *cols: ColumnInputType) -> DataFrame:
         """Perform grouped sum on this GroupedDataFrame.
 
@@ -886,7 +1097,30 @@ class GroupedDataFrame:
 
         return self.df._agg([(c, "mean") for c in cols], group_by=self.group_by)
 
-    def agg(self, to_agg: List[Tuple[ColumnInputType, str]]) -> DataFrame:
+    def min(self, *cols: ColumnInputType) -> DataFrame:
+        """Perform grouped min on this GroupedDataFrame.
+
+        Args:
+            *cols (Union[str, Expression]): columns to min
+
+        Returns:
+            DataFrame: DataFrame with grouped min.
+        """
+        return self.df._agg([(c, "min") for c in cols], group_by=self.group_by)
+
+    def max(self, *cols: ColumnInputType) -> DataFrame:
+        """Performs grouped max on this GroupedDataFrame.
+
+        Args:
+            *cols (Union[str, Expression]): columns to max
+
+        Returns:
+            DataFrame: DataFrame with grouped max.
+        """
+
+        return self.df._agg([(c, "max") for c in cols], group_by=self.group_by)
+
+    def agg(self, to_agg: list[tuple[ColumnInputType, str]]) -> DataFrame:
         """Perform aggregations on this GroupedDataFrame. Allows for mixed aggregations.
 
         Example:
