@@ -1,7 +1,8 @@
+from __future__ import annotations
+
 import sys
 from dataclasses import dataclass
 from enum import Enum
-from typing import Any, Dict, List
 
 if sys.version_info < (3, 8):
     from typing_extensions import Protocol
@@ -29,7 +30,7 @@ class SourceInfo(Protocol):
 @dataclass(frozen=True)
 class CSVSourceInfo(SourceInfo):
 
-    filepaths: List[str]
+    filepaths: list[str]
     delimiter: str
     has_headers: bool
 
@@ -43,7 +44,7 @@ class CSVSourceInfo(SourceInfo):
 @dataclass(frozen=True)
 class JSONSourceInfo(SourceInfo):
 
-    filepaths: List[str]
+    filepaths: list[str]
 
     def scan_type(self):
         return StorageType.JSON
@@ -52,23 +53,10 @@ class JSONSourceInfo(SourceInfo):
         return len(self.filepaths)
 
 
-@dataclass(frozen=True)
-class InMemorySourceInfo(SourceInfo):
-
-    data: Dict[str, List[Any]]
-    num_partitions: int = 1
-
-    def scan_type(self):
-        return StorageType.IN_MEMORY
-
-    def get_num_partitions(self) -> int:
-        return self.num_partitions
-
-
 @dataclass(frozen=True)
 class ParquetSourceInfo(SourceInfo):
 
-    filepaths: List[str]
+    filepaths: list[str]
 
     def scan_type(self):
         return StorageType.PARQUET
