@@ -5,7 +5,7 @@ from dataclasses import dataclass
 from enum import Enum
 from functools import lru_cache, partial
 from types import MappingProxyType
-from typing import Callable, FrozenSet, Optional, Tuple, TypeVar
+from typing import Callable, FrozenSet, Tuple, TypeVar
 
 from daft.types import ExpressionType
 
@@ -23,7 +23,7 @@ class ExpressionOperator:
     name: str
     nargs: int
     type_matrix: TypeMatrix
-    symbol: Optional[str] = None
+    symbol: str | None = None
 
     def __post_init__(self) -> None:
         for k, v in self.type_matrix:
@@ -34,10 +34,10 @@ class ExpressionOperator:
             assert isinstance(v, ExpressionType), f"{v} is not an ExpressionType"
 
     @lru_cache(maxsize=1)
-    def _type_matrix_dict(self) -> MappingProxyType[Tuple[ExpressionType, ...], ExpressionType]:
+    def _type_matrix_dict(self) -> MappingProxyType[tuple[ExpressionType, ...], ExpressionType]:
         return MappingProxyType(dict(self.type_matrix))
 
-    def get_return_type(self, args: Tuple[ExpressionType, ...]) -> Optional[ExpressionType]:
+    def get_return_type(self, args: tuple[ExpressionType, ...]) -> ExpressionType | None:
         # Treat all Python types as a PY[object] for the purposes of typing
         args = tuple([ExpressionType.python_object() if ExpressionType.is_py(a) else a for a in args])
 
@@ -73,6 +73,17 @@ _BinaryNumericalTM = frozenset(
     }.items()
 )
 
+_ComparableTM = frozenset(
+    {
+        (ExpressionType.integer(),): ExpressionType.integer(),
+        (ExpressionType.logical(),): ExpressionType.logical(),
+        (ExpressionType.float(),): ExpressionType.float(),
+        (ExpressionType.string(),): ExpressionType.string(),
+        (ExpressionType.date(),): ExpressionType.date(),
+        (ExpressionType.bytes(),): ExpressionType.bytes(),
+    }.items()
+)
+
 _ComparisionTM = frozenset(
     {
         (ExpressionType.integer(), ExpressionType.integer()): ExpressionType.logical(),
@@ -85,6 +96,7 @@ _ComparisionTM = frozenset(
     }.items()
 )
 
+
 _BinaryLogicalTM = frozenset(
     {
         (ExpressionType.logical(), ExpressionType.logical()): ExpressionType.logical(),
@@ -144,7 +156,13 @@ _DatetimeExtractionTM = frozenset(
     }.items()
 )
 
-_LengthTM = frozenset(
+_StrConcatTM = frozenset(
+    {
+        (ExpressionType.string(), ExpressionType.string()): ExpressionType.string(),
+    }.items()
+)
+
+_StrLengthTM = frozenset(
     {
         (ExpressionType.string(),): ExpressionType.integer(),
     }.items()
@@ -168,6 +186,8 @@ _NBop = partial(_BOp, type_matrix=_BinaryNumericalTM)
 # Comparison Binary Ops
 _CBop = partial(_BOp, type_matrix=_ComparisionTM)
 
+_ComparibleUop = partial(_UOp, type_matrix=_ComparableTM)
+
 # Logical Binary Ops
 _LBop = partial(_BOp, type_matrix=_BinaryLogicalTM)
 
@@ -188,19 +208,82 @@ class OperatorEnum(Enum):
     # Reductions
     SUM = _NUop(name="sum", symbol="sum")
     MEAN = _NUop(name="mean", symbol="mean")
-    MIN = _NUop(name="min", symbol="min")
-    MAX = _NUop(name="max", symbol="max")
+    MIN = _ComparibleUop(name="min", symbol="min")
+    MAX = _ComparibleUop(name="max", symbol="max")
 
     COUNT = _UOp(name="count", symbol="count", type_matrix=_CountLogicalTM)
 
     # Logical
     INVERT = _UOp(name="invert", symbol="~", type_matrix=_UnaryLogicalTM)
 
+    # Cast
+    CAST_INT = _UOp(
+        name="cast_int",
+        type_matrix=frozenset(
+            {
+                (ExpressionType.integer(),): ExpressionType.integer(),
+                (ExpressionType.float(),): ExpressionType.integer(),
+                (ExpressionType.logical(),): ExpressionType.integer(),
+                (ExpressionType.python_object(),): ExpressionType.integer(),
+            }.items()
+        ),
+    )
+    CAST_FLOAT = _UOp(
+        name="cast_float",
+        type_matrix=frozenset(
+            {
+                (ExpressionType.float(),): ExpressionType.float(),
+                (ExpressionType.integer(),): ExpressionType.float(),
+                (ExpressionType.python_object(),): ExpressionType.float(),
+            }.items()
+        ),
+    )
+    CAST_STRING = _UOp(
+        name="cast_string",
+        type_matrix=frozenset(
+            {
+                (ExpressionType.string(),): ExpressionType.string(),
+                (ExpressionType.integer(),): ExpressionType.string(),
+                (ExpressionType.logical(),): ExpressionType.string(),
+                (ExpressionType.float(),): ExpressionType.string(),
+                (ExpressionType.python_object(),): ExpressionType.string(),
+            }.items()
+        ),
+    )
+    CAST_LOGICAL = _UOp(
+        name="cast_logical",
+        type_matrix=frozenset(
+            {
+                (ExpressionType.logical(),): ExpressionType.logical(),
+                (ExpressionType.python_object(),): ExpressionType.logical(),
+            }.items()
+        ),
+    )
+    CAST_DATE = _UOp(
+        name="cast_date",
+        type_matrix=frozenset(
+            {
+                (ExpressionType.date(),): ExpressionType.date(),
+                (ExpressionType.python_object(),): ExpressionType.date(),
+            }.items()
+        ),
+    )
+    CAST_BYTES = _UOp(
+        name="cast_bytes",
+        type_matrix=frozenset(
+            {
+                (ExpressionType.bytes(),): ExpressionType.bytes(),
+                (ExpressionType.python_object(),): ExpressionType.bytes(),
+            }.items()
+        ),
+    )
+
     # String
+    STR_CONCAT = _BOp(name="str_concat", symbol="concat", type_matrix=_StrConcatTM)
     STR_CONTAINS = _LSop(name="str_contains", symbol="contains")
     STR_ENDSWITH = _LSop(name="str_endswith", symbol="endswith")
     STR_STARTSWITH = _LSop(name="str_startswith", symbol="startswith")
-    STR_LENGTH = _UOp(name="str_length", symbol="len", type_matrix=_LengthTM)
+    STR_LENGTH = _UOp(name="str_length", symbol="len", type_matrix=_StrLengthTM)
 
     # Null
     IS_NULL = _UOp(name="is_null", symbol="is_null", type_matrix=_AllLogicalTM)
@@ -261,6 +344,13 @@ class OperatorEvaluator(Protocol[ValueType]):
     COUNT: UnaryFunction
     INVERT: UnaryFunction
 
+    CAST_INT: UnaryFunction
+    CAST_FLOAT: UnaryFunction
+    CAST_STRING: UnaryFunction
+    CAST_LOGICAL: UnaryFunction
+    CAST_DATE: UnaryFunction
+    CAST_BYTES: UnaryFunction
+
     ADD: BinaryFunction
     SUB: BinaryFunction
     MUL: BinaryFunction
@@ -277,6 +367,7 @@ class OperatorEvaluator(Protocol[ValueType]):
     GT: BinaryFunction
     GE: BinaryFunction
 
+    STR_CONCAT: BinaryFunction
     STR_CONTAINS: BinaryFunction
     STR_ENDSWITH: BinaryFunction
     STR_STARTSWITH: BinaryFunction
