@@ -1,5 +1,7 @@
+from __future__ import annotations
+
 import pathlib
-from typing import Dict, List, Optional
+from urllib.parse import urlparse
 
 import pyarrow as pa
 from loguru import logger
@@ -8,7 +10,7 @@ from daft import filesystem
 from daft.runners.blocks import ArrowDataBlock
 
 
-def download(url_block: ArrowDataBlock) -> List[Optional[bytes]]:
+def download(url_block: ArrowDataBlock) -> list[bytes | None]:
     assert isinstance(
         url_block, ArrowDataBlock
     ), f"Can only download from columns containing strings, found non-arrow block"
@@ -16,18 +18,22 @@ def download(url_block: ArrowDataBlock) -> List[Optional[bytes]]:
         url_block.data.type
     ), f"Can only download from columns containing strings, found {url_block.data.type}"
 
-    results: List[Optional[bytes]] = [None for _ in range(len(url_block))]
+    results: list[bytes | None] = [None for _ in range(len(url_block))]
 
     path_to_result_idx = {}
-    to_download: Dict[str, List[str]] = {}
+    to_download: dict[str, list[str]] = {}
     for i, path in enumerate(url_block.iter_py()):
         if path is None:
             continue
         protocol = filesystem.get_protocol_from_path(path)
 
-        # fsspec returns data keyed by absolute paths, so we do a conversion here to match up later
+        # fsspec returns data keyed by absolute paths
         if protocol == "file":
             path = str(pathlib.Path(path).resolve())
+        # fsspec strips the s3:// scheme from the path after .cat
+        elif protocol == "s3":
+            parsed_url = urlparse(path)
+            path = parsed_url.netloc + parsed_url.path
 
         if protocol not in to_download:
             to_download[protocol] = []
