@@ -1,5 +1,6 @@
 from __future__ import annotations
 
+import datetime
 import itertools
 import warnings
 from abc import abstractmethod
@@ -8,15 +9,12 @@ from functools import partial, partialmethod
 from typing import (
     Any,
     Callable,
-    Dict,
-    List,
     NewType,
-    Optional,
     Sequence,
     Tuple,
-    Type,
     TypeVar,
     cast,
+    get_type_hints,
 )
 
 import numpy as np
@@ -73,7 +71,7 @@ ColID = NewType("ColID", int)
 
 
 class ExpressionExecutor:
-    def eval(self, expr: Expression, operands: Dict[str, Any]) -> DataBlock:
+    def eval(self, expr: Expression, operands: dict[str, Any]) -> DataBlock:
         result: DataBlock
         if isinstance(expr, ColumnExpression):
             name = expr.name()
@@ -87,7 +85,7 @@ class ExpressionExecutor:
             result = self.eval(expr._expr, operands)
             return result
         elif isinstance(expr, CallExpression):
-            eval_args: Tuple[DataBlock, ...] = tuple(self.eval(a, operands) for a in expr._args)
+            eval_args: tuple[DataBlock, ...] = tuple(self.eval(a, operands) for a in expr._args)
 
             # Use a PyListDataBlock evaluator if any of the args are Python types
             op_evaluator = (
@@ -114,7 +112,7 @@ class ExpressionExecutor:
 class Expression(TreeNode["Expression"]):
     def __init__(self) -> None:
         super().__init__()
-        self._id: Optional[ColID] = None
+        self._id: ColID | None = None
 
     def __repr__(self) -> str:
         if self.has_id():
@@ -129,10 +127,8 @@ class Expression(TreeNode["Expression"]):
 
     def __bool__(self) -> bool:
         raise ValueError(
-            (
-                "Expressions don't have a truth value until executed. "
-                "If you reached this error using `and` / `or`, use `&` / `|` instead."
-            )
+            "Expressions don't have a truth value until executed. "
+            "If you reached this error using `and` / `or`, use `&` / `|` instead."
         )
 
     def _unary_op(
@@ -168,11 +164,11 @@ class Expression(TreeNode["Expression"]):
         )
 
     @abstractmethod
-    def resolved_type(self) -> Optional[ExpressionType]:
+    def resolved_type(self) -> ExpressionType | None:
         """Expressions have a resolved_type only if they are resolved"""
         return None
 
-    def name(self) -> Optional[str]:
+    def name(self) -> str | None:
         for child in self._children():
             name = child.name()
             if name is not None:
@@ -190,7 +186,7 @@ class Expression(TreeNode["Expression"]):
                 assert self._id is not None
                 return self._id
 
-    def get_id(self) -> Optional[ColID]:
+    def get_id(self) -> ColID | None:
         return self._id
 
     def has_id(self) -> bool:
@@ -206,8 +202,8 @@ class Expression(TreeNode["Expression"]):
         ce.resolve_to_expression(self)
         return ce
 
-    def required_columns(self, unresolved_only: bool = False) -> List[ColumnExpression]:
-        to_rtn: List[ColumnExpression] = []
+    def required_columns(self, unresolved_only: bool = False) -> list[ColumnExpression]:
+        to_rtn: list[ColumnExpression] = []
         for child in self._children():
             to_rtn.extend(child.required_columns(unresolved_only))
         return to_rtn
@@ -304,7 +300,7 @@ class Expression(TreeNode["Expression"]):
         """
         return AliasExpression(self, name)
 
-    def as_py(self, type_: Type) -> AsPyExpression:
+    def as_py(self, type_: type) -> AsPyExpression:
         """Treats every value on the given expression as an object of the specified type. Users can then call
         methods on this object, which will be translated into a method call on every value on the Expression.
 
@@ -323,7 +319,7 @@ class Expression(TreeNode["Expression"]):
         """
         return AsPyExpression(self, ExpressionType.from_py_type(type_))
 
-    def apply(self, func: Callable, return_type: Optional[Type] = None) -> Expression:
+    def apply(self, func: Callable, return_type: type | None = None) -> Expression:
         """Apply a function on a given expression
 
         Example:
@@ -334,15 +330,18 @@ class Expression(TreeNode["Expression"]):
 
         Args:
             func (Callable): Function to run per value of the expression
-            return_type (Optional[Type], optional): Return type of the function that was ran. Defaults to None.
+            return_type (Optional[Type], optional): Return type of the function that was ran. This defaults to None and Daft will infer the return_type
+                from the function's type annotations if available.
 
         Returns:
             Expression: New expression after having run the function on the expression
         """
+        inferred_type = get_type_hints(func).get("return", None)
+        return_type = inferred_type if inferred_type is not None else return_type
         if return_type is None:
             warnings.warn(
-                "No `return_type` keyword argument specified in `.apply`. It is highly recommended to "
-                "specify a return_type for Daft to perform optimizations and for access to vectorized expression operators."
+                f"Supplied function {func} was not annotated with a return type and no `return_type` keyword argument specified in `.apply`."
+                "It is highly recommended to specify a return_type for Daft to perform optimizations and for access to vectorized expression operators."
             )
 
         expression_type = (
@@ -445,6 +444,60 @@ class Expression(TreeNode["Expression"]):
             (self,),
         )
 
+    def cast(self, to: type) -> Expression:
+        """Casts an expression to a given type
+
+        Casting defaults to a set of "reasonable behaviors" for each ``(from_type, to_type)`` pair. For more fine-grained control, please consult
+        each type's method accessors or make a feature request for an accessor if none exists for your use-case.
+
+        This method supports:
+
+        1. Casting of a PY type to a Primitive type (e.g. converting a PY[object] column of strings to STRING with ``.cast(str)``) - this will coerce each Python object into the specified primitive type, and then convert the data to an optimized backend representation.
+        2. Casting between Primitive types, with a set of reasonable default behavior for many type pairs
+
+        Example:
+
+            >>> # [1.0, 2.5, None]: FLOAT -> [1, 2, None]: INTEGER
+            >>> col("float").cast(int)
+            >>>
+            >>> # [Path("/tmp1"), Path("/tmp2"), Path("/tmp3")]: PY[Path] -> ["/tmp1", "/tmp1", "/tmp1"]: STRING
+            >>> col("path_obj_col").cast(str)
+
+        Returns:
+            Expression: Expression with the specified new type
+        """
+        if to == str:
+            return CallExpression(
+                OperatorEnum.CAST_STRING,
+                (self,),
+            )
+        elif to == int:
+            return CallExpression(
+                OperatorEnum.CAST_INT,
+                (self,),
+            )
+        elif to == float:
+            return CallExpression(
+                OperatorEnum.CAST_FLOAT,
+                (self,),
+            )
+        elif to == bool:
+            return CallExpression(
+                OperatorEnum.CAST_LOGICAL,
+                (self,),
+            )
+        elif to == datetime.date:
+            return CallExpression(
+                OperatorEnum.CAST_DATE,
+                (self,),
+            )
+        elif to == bytes:
+            return CallExpression(
+                OperatorEnum.CAST_BYTES,
+                (self,),
+            )
+        raise NotImplementedError(f"Casting to a non-primitive Python object {to} not yet implemented")
+
     ###
     # Accessors
     ###
@@ -470,7 +523,7 @@ class LiteralExpression(Expression):
         super().__init__()
         self._value = value
 
-    def resolved_type(self) -> Optional[ExpressionType]:
+    def resolved_type(self) -> ExpressionType | None:
         return ExpressionType.from_py_type(type(self._value))
 
     def _display_str(self) -> str:
@@ -484,13 +537,13 @@ class CallExpression(Expression):
     def __init__(
         self,
         operator: OperatorEnum,
-        func_args: Tuple,
+        func_args: tuple,
     ) -> None:
         super().__init__()
         self._args_ids = tuple(self._register_child(self._to_expression(arg)) for arg in func_args)
         self._operator = operator
 
-    def resolved_type(self) -> Optional[ExpressionType]:
+    def resolved_type(self) -> ExpressionType | None:
         args_resolved_types = tuple(arg.resolved_type() for arg in self._args)
         if any([arg_type is None for arg_type in args_resolved_types]):
             return None
@@ -510,7 +563,7 @@ class CallExpression(Expression):
         return ret_type
 
     @property
-    def _args(self) -> Tuple[Expression, ...]:
+    def _args(self) -> tuple[Expression, ...]:
         return tuple(self._children()[i] for i in self._args_ids)
 
     def _display_str(self) -> str:
@@ -536,9 +589,9 @@ class UdfExpression(Expression):
         self,
         func: Callable[..., Sequence],
         func_ret_type: ExpressionType,
-        func_args: Tuple,
-        func_kwargs: Dict[str, Any],
-        resource_request: Optional[ResourceRequest] = None,
+        func_args: tuple,
+        func_kwargs: dict[str, Any],
+        resource_request: ResourceRequest | None = None,
     ) -> None:
         super().__init__()
         self._func = func
@@ -553,14 +606,14 @@ class UdfExpression(Expression):
         return ResourceRequest.default()
 
     @property
-    def _args(self) -> Tuple[Expression, ...]:
+    def _args(self) -> tuple[Expression, ...]:
         return tuple(self._children()[i] for i in self._args_ids)
 
     @property
-    def _kwargs(self) -> Dict[str, Expression]:
+    def _kwargs(self) -> dict[str, Expression]:
         return {kw: self._children()[i] for kw, i in self._kwargs_ids.items()}
 
-    def resolved_type(self) -> Optional[ExpressionType]:
+    def resolved_type(self) -> ExpressionType | None:
         return self._func_ret_type
 
     def _display_str(self) -> str:
@@ -618,14 +671,14 @@ T = TypeVar("T")
 
 
 class ColumnExpression(Expression):
-    def __init__(self, name: str, expr_type: Optional[ExpressionType] = None) -> None:
+    def __init__(self, name: str, expr_type: ExpressionType | None = None) -> None:
         super().__init__()
         if not isinstance(name, str):
             raise TypeError(f"Expected name to be type str, is {type(name)}")
         self._name = name
         self._type = expr_type
 
-    def resolved_type(self) -> Optional[ExpressionType]:
+    def resolved_type(self) -> ExpressionType | None:
         return self._type
 
     def _display_str(self) -> str:
@@ -640,10 +693,10 @@ class ColumnExpression(Expression):
     def __repr__(self) -> str:
         return self._display_str()
 
-    def name(self) -> Optional[str]:
+    def name(self) -> str | None:
         return self._name
 
-    def required_columns(self, unresolved_only: bool = False) -> List[ColumnExpression]:
+    def required_columns(self, unresolved_only: bool = False) -> list[ColumnExpression]:
         if unresolved_only and self.has_id():
             return []
         return [self]
@@ -668,7 +721,7 @@ class AliasExpression(Expression):
         self._register_child(expr)
         self._name = name
 
-    def resolved_type(self) -> Optional[ExpressionType]:
+    def resolved_type(self) -> ExpressionType | None:
         return self._expr.resolved_type()
 
     @property
@@ -678,10 +731,10 @@ class AliasExpression(Expression):
     def _display_str(self) -> str:
         return f"[{self._expr}].alias({self._name})"
 
-    def name(self) -> Optional[str]:
+    def name(self) -> str | None:
         return self._name
 
-    def get_id(self) -> Optional[ColID]:
+    def get_id(self) -> ColID | None:
         return self._expr.get_id()
 
     def _assign_id(self, strict: bool = True) -> ColID:
@@ -696,7 +749,7 @@ class AsPyExpression(Expression):
         self,
         expr: Expression,
         type_: ExpressionType,
-        attr_name: Optional[str] = "__call__",
+        attr_name: str | None = "__call__",
     ) -> None:
         super().__init__()
         self._type = type_
@@ -759,10 +812,10 @@ class AsPyExpression(Expression):
     def _expr(self) -> Expression:
         return self._children()[0]
 
-    def resolved_type(self) -> Optional[ExpressionType]:
+    def resolved_type(self) -> ExpressionType | None:
         return self._type
 
-    def get_id(self) -> Optional[ColID]:
+    def get_id(self) -> ColID | None:
         return self._expr.get_id()
 
     def _assign_id(self, strict: bool = True) -> ColID:
@@ -794,6 +847,23 @@ class UrlMethodAccessor(BaseMethodAccessor):
 
 
 class StringMethodAccessor(BaseMethodAccessor):
+    def concat(self, other: Any) -> CallExpression:
+        """Concatenates two string expressions together
+
+        Args:
+            other (Expression): a string expression to concatenate with
+
+        Returns:
+            CallExpression: a STRING expression which is `self` concatenated with `other`
+        """
+        if isinstance(other, str):
+            other = lit(other)
+        assert isinstance(other, Expression)
+        return CallExpression(
+            OperatorEnum.STR_CONCAT,
+            (self._expr, other),
+        )
+
     def contains(self, pattern: str) -> CallExpression:
         """Checks whether each string contains the given pattern in a string column
 
