@@ -24,7 +24,7 @@ namespace bit_util = arrow::bit_util;
 #endif
 
 template <typename InType>
-struct SearchSortedPrimativeSingle {
+struct SearchSortedPrimitiveSingle {
   static void Exec(const arrow::ArrayData *arr, const arrow::ArrayData *keys, arrow::ArrayData *result, const bool input_reversed) {
     if (keys->GetNullCount() == 0) {
       return KernelNonNull(arr, keys, result, input_reversed);
@@ -106,7 +106,7 @@ struct SearchSortedPrimativeSingle {
     const T *keys_ptr = keys->GetValues<T>(1);
     ARROW_CHECK(keys_ptr != NULL);
 
-    const uint8_t *keys_bitmask_ptr = keys->GetValues<uint8_t>(0);
+    const uint8_t *keys_bitmask_ptr = keys->GetValues<uint8_t>(0, 0);
     ARROW_CHECK(keys_bitmask_ptr != NULL);
 
     ARROW_CHECK(result->type->id() == arrow::Type::UINT64);
@@ -114,35 +114,36 @@ struct SearchSortedPrimativeSingle {
 
     uint64_t *result_ptr = result->GetMutableValues<uint64_t>(1);
     ARROW_CHECK(result_ptr != NULL);
+    const size_t result_offset = result->offset;
 
-    uint8_t *result_bitmask_ptr = result->GetMutableValues<uint8_t>(0);
-    ARROW_CHECK(result_bitmask_ptr != NULL);
+    uint8_t *result_bitmask_ptr = result->GetMutableValues<uint8_t>(0, 0);
+    ARROW_CHECK(result_bitmask_ptr == NULL);
 
     auto cmp = std::less<T>{};
     size_t min_idx = 0;
     size_t arr_len = arr->length;
+
     size_t max_idx = arr_len;
+
     size_t key_len = keys->length;
+    const size_t key_offset = keys->offset;
 
     if (key_len == 0) {
       return;
     }
 
     T last_key_val = *keys_ptr;
+    bool last_key_is_valid = bit_util::GetBit(keys_bitmask_ptr, key_offset);
 
     for (size_t key_idx = 0; key_idx < key_len; key_idx++, keys_ptr++, result_ptr++) {
-      const bool key_bit = bit_util::GetBit(keys_bitmask_ptr, key_idx);
-      bit_util::SetBitTo(result_bitmask_ptr, key_idx, key_bit);
-      if (!key_bit) {
-        continue;
-      }
+      const bool curr_key_is_valid = bit_util::GetBit(keys_bitmask_ptr, key_idx + key_offset);
       const T key_val = *keys_ptr;
       /*
        * Updating only one of the indices based on the previous key
        * gives the search a big boost when keys are sorted, but slightly
        * slows down things for purely random ones.
        */
-      if (cmp(last_key_val, key_val)) {
+      if ((!curr_key_is_valid) || (last_key_is_valid && cmp(last_key_val, key_val))) {
         max_idx = arr_len;
       } else {
         min_idx = 0;
@@ -150,12 +151,13 @@ struct SearchSortedPrimativeSingle {
       }
 
       last_key_val = key_val;
+      last_key_is_valid = curr_key_is_valid;
 
       while (min_idx < max_idx) {
         const size_t mid_idx = min_idx + ((max_idx - min_idx) >> 1);
         const size_t corrected_idx = input_reversed ? arr_len - mid_idx - 1 : mid_idx;
         const T mid_val = *(arr_ptr + corrected_idx);
-        if (cmp(mid_val, key_val)) {
+        if ((!curr_key_is_valid) || cmp(mid_val, key_val)) {
           min_idx = mid_idx + 1;
         } else {
           max_idx = mid_idx;
@@ -166,48 +168,48 @@ struct SearchSortedPrimativeSingle {
   }
 };
 
-void search_sorted_primative_single(const arrow::ArrayData *arr, const arrow::ArrayData *keys, arrow::ArrayData *result,
+void search_sorted_primitive_single(const arrow::ArrayData *arr, const arrow::ArrayData *keys, arrow::ArrayData *result,
                                     const bool input_reversed) {
   switch (arr->type->id()) {
     case arrow::Type::INT8:
-      return SearchSortedPrimativeSingle<arrow::Int8Type>::Exec(arr, keys, result, input_reversed);
+      return SearchSortedPrimitiveSingle<arrow::Int8Type>::Exec(arr, keys, result, input_reversed);
     case arrow::Type::INT16:
-      return SearchSortedPrimativeSingle<arrow::Int16Type>::Exec(arr, keys, result, input_reversed);
+      return SearchSortedPrimitiveSingle<arrow::Int16Type>::Exec(arr, keys, result, input_reversed);
     case arrow::Type::INT32:
-      return SearchSortedPrimativeSingle<arrow::Int32Type>::Exec(arr, keys, result, input_reversed);
+      return SearchSortedPrimitiveSingle<arrow::Int32Type>::Exec(arr, keys, result, input_reversed);
     case arrow::Type::INT64:
-      return SearchSortedPrimativeSingle<arrow::Int64Type>::Exec(arr, keys, result, input_reversed);
+      return SearchSortedPrimitiveSingle<arrow::Int64Type>::Exec(arr, keys, result, input_reversed);
     case arrow::Type::UINT8:
-      return SearchSortedPrimativeSingle<arrow::UInt8Type>::Exec(arr, keys, result, input_reversed);
+      return SearchSortedPrimitiveSingle<arrow::UInt8Type>::Exec(arr, keys, result, input_reversed);
     case arrow::Type::UINT16:
-      return SearchSortedPrimativeSingle<arrow::UInt16Type>::Exec(arr, keys, result, input_reversed);
+      return SearchSortedPrimitiveSingle<arrow::UInt16Type>::Exec(arr, keys, result, input_reversed);
     case arrow::Type::UINT32:
-      return SearchSortedPrimativeSingle<arrow::UInt32Type>::Exec(arr, keys, result, input_reversed);
+      return SearchSortedPrimitiveSingle<arrow::UInt32Type>::Exec(arr, keys, result, input_reversed);
     case arrow::Type::UINT64:
-      return SearchSortedPrimativeSingle<arrow::UInt64Type>::Exec(arr, keys, result, input_reversed);
+      return SearchSortedPrimitiveSingle<arrow::UInt64Type>::Exec(arr, keys, result, input_reversed);
     case arrow::Type::FLOAT:
-      return SearchSortedPrimativeSingle<arrow::FloatType>::Exec(arr, keys, result, input_reversed);
+      return SearchSortedPrimitiveSingle<arrow::FloatType>::Exec(arr, keys, result, input_reversed);
     case arrow::Type::DOUBLE:
-      return SearchSortedPrimativeSingle<arrow::DoubleType>::Exec(arr, keys, result, input_reversed);
+      return SearchSortedPrimitiveSingle<arrow::DoubleType>::Exec(arr, keys, result, input_reversed);
     case arrow::Type::DATE32:
-      return SearchSortedPrimativeSingle<arrow::Date32Type>::Exec(arr, keys, result, input_reversed);
+      return SearchSortedPrimitiveSingle<arrow::Date32Type>::Exec(arr, keys, result, input_reversed);
     case arrow::Type::DATE64:
-      return SearchSortedPrimativeSingle<arrow::Date64Type>::Exec(arr, keys, result, input_reversed);
+      return SearchSortedPrimitiveSingle<arrow::Date64Type>::Exec(arr, keys, result, input_reversed);
     case arrow::Type::TIME32:
-      return SearchSortedPrimativeSingle<arrow::Time32Type>::Exec(arr, keys, result, input_reversed);
+      return SearchSortedPrimitiveSingle<arrow::Time32Type>::Exec(arr, keys, result, input_reversed);
     case arrow::Type::TIME64:
-      return SearchSortedPrimativeSingle<arrow::Time64Type>::Exec(arr, keys, result, input_reversed);
+      return SearchSortedPrimitiveSingle<arrow::Time64Type>::Exec(arr, keys, result, input_reversed);
     case arrow::Type::TIMESTAMP:
-      return SearchSortedPrimativeSingle<arrow::TimestampType>::Exec(arr, keys, result, input_reversed);
+      return SearchSortedPrimitiveSingle<arrow::TimestampType>::Exec(arr, keys, result, input_reversed);
     case arrow::Type::DURATION:
-      return SearchSortedPrimativeSingle<arrow::DurationType>::Exec(arr, keys, result, input_reversed);
+      return SearchSortedPrimitiveSingle<arrow::DurationType>::Exec(arr, keys, result, input_reversed);
     case arrow::Type::INTERVAL_MONTHS:
-      return SearchSortedPrimativeSingle<arrow::MonthIntervalType>::Exec(arr, keys, result, input_reversed);
+      return SearchSortedPrimitiveSingle<arrow::MonthIntervalType>::Exec(arr, keys, result, input_reversed);
     // Need custom less function for this since it uses a custom struct for the data structure
     // case arrow::Type::INTERVAL_MONTH_DAY_NANO:
-    //   return SearchSortedPrimativeSingle<arrow::MonthDayNanoIntervalType>::Exec(arr, keys, result);
+    //   return SearchSortedPrimitiveSingle<arrow::MonthDayNanoIntervalType>::Exec(arr, keys, result);
     case arrow::Type::INTERVAL_DAY_TIME:
-      return SearchSortedPrimativeSingle<arrow::DayTimeIntervalType>::Exec(arr, keys, result, input_reversed);
+      return SearchSortedPrimitiveSingle<arrow::DayTimeIntervalType>::Exec(arr, keys, result, input_reversed);
     default:
       ARROW_LOG(FATAL) << "Unknown arrow type" << arr->type->id();
   }
@@ -226,17 +228,17 @@ void search_sorted_binary_single(const arrow::ArrayData *arr, const arrow::Array
   const T *arr_index_ptr = arr->GetValues<T>(1);
   ARROW_CHECK(arr_index_ptr != NULL);
 
-  const uint8_t *arr_data_ptr = arr->GetValues<uint8_t>(2);
+  const uint8_t *arr_data_ptr = arr->GetValues<uint8_t>(2, 0);
   ARROW_CHECK(arr_data_ptr != NULL);
 
   const bool has_nulls = keys->GetNullCount() > 0;
-  const uint8_t *keys_bitmask_ptr = keys->GetValues<uint8_t>(0);
+  const uint8_t *keys_bitmask_ptr = keys->GetValues<uint8_t>(0, 0);
   // ARROW_CHECK(keys_bitmask_ptr != NULL);
 
   const T *keys_index_ptr = keys->GetValues<T>(1);
   ARROW_CHECK(keys_index_ptr != NULL);
 
-  const uint8_t *keys_data_ptr = keys->GetValues<uint8_t>(2);
+  const uint8_t *keys_data_ptr = keys->GetValues<uint8_t>(2, 0);
   ARROW_CHECK(keys_data_ptr != NULL);
 
   ARROW_CHECK(result->length == keys->length);
@@ -244,26 +246,28 @@ void search_sorted_binary_single(const arrow::ArrayData *arr, const arrow::Array
   uint64_t *result_ptr = result->GetMutableValues<uint64_t>(1);
   ARROW_CHECK(result_ptr != NULL);
 
-  uint8_t *result_bitmask_ptr = result->GetMutableValues<uint8_t>(0);
-  ARROW_CHECK(!has_nulls || (result_bitmask_ptr != NULL));
+  uint8_t *result_bitmask_ptr = result->GetMutableValues<uint8_t>(0, 0);
+  ARROW_CHECK(result_bitmask_ptr == NULL);
 
   size_t min_idx = 0;
   size_t arr_len = arr->length;
   size_t max_idx = arr_len;
   size_t key_len = keys->length;
+  const size_t key_offset = keys->offset;
+
   size_t last_key_idx = 0;
 
   if (key_len == 0) {
     return;
   }
-
+  bool last_key_is_valid = true;
+  if (has_nulls) {
+    last_key_is_valid = bit_util::GetBit(keys_bitmask_ptr, key_offset);
+  }
   for (size_t key_idx = 0; key_idx < key_len; key_idx++, result_ptr++) {
+    bool key_is_valid = true;
     if (has_nulls) {
-      const bool key_bit = bit_util::GetBit(keys_bitmask_ptr, key_idx);
-      bit_util::SetBitTo(result_bitmask_ptr, key_idx, key_bit);
-      if (!key_bit) {
-        continue;
-      }
+      key_is_valid = bit_util::GetBit(keys_bitmask_ptr, key_idx + key_offset);
     }
     const size_t curr_key_offset = keys_index_ptr[key_idx];
     const size_t curr_key_size = keys_index_ptr[key_idx + 1] - curr_key_offset;
@@ -275,8 +279,9 @@ void search_sorted_binary_single(const arrow::ArrayData *arr, const arrow::Array
      * gives the search a big boost when keys are sorted, but slightly
      * slows down things for purely random ones.
      */
-    if (daft::kernels::strcmpNoTerminator(keys_data_ptr + last_key_offset, keys_data_ptr + curr_key_offset, last_key_size, curr_key_size) <
-        0) {
+    if ((!key_is_valid) ||
+        (last_key_is_valid && (daft::kernels::strcmpNoTerminator(keys_data_ptr + last_key_offset, keys_data_ptr + curr_key_offset,
+                                                                 last_key_size, curr_key_size) < 0))) {
       max_idx = arr_len;
     } else {
       min_idx = 0;
@@ -284,6 +289,7 @@ void search_sorted_binary_single(const arrow::ArrayData *arr, const arrow::Array
     }
 
     last_key_idx = key_idx;
+    last_key_is_valid = key_is_valid;
 
     while (min_idx < max_idx) {
       const size_t mid_idx = min_idx + ((max_idx - min_idx) >> 1);
@@ -291,8 +297,8 @@ void search_sorted_binary_single(const arrow::ArrayData *arr, const arrow::Array
       const size_t mid_arr_offset = arr_index_ptr[corrected_idx];
       const size_t mid_arr_size = arr_index_ptr[corrected_idx + 1] - mid_arr_offset;
 
-      if (daft::kernels::strcmpNoTerminator(arr_data_ptr + mid_arr_offset, keys_data_ptr + curr_key_offset, mid_arr_size, curr_key_size) <
-          0) {
+      if ((!key_is_valid) || (daft::kernels::strcmpNoTerminator(arr_data_ptr + mid_arr_offset, keys_data_ptr + curr_key_offset,
+                                                                mid_arr_size, curr_key_size) < 0)) {
         min_idx = mid_idx + 1;
       } else {
         max_idx = mid_idx;
@@ -302,48 +308,48 @@ void search_sorted_binary_single(const arrow::ArrayData *arr, const arrow::Array
   }
 }
 
-void add_primative_memory_view_to_comp_view(daft::kernels::CompositeView &comp_view, const std::shared_ptr<arrow::ArrayData> arr) {
+void add_primitive_memory_view_to_comp_view(daft::kernels::CompositeView &comp_view, const std::shared_ptr<arrow::ArrayData> arr) {
   ARROW_CHECK(arrow::is_primitive(arr->type->id()));
   switch (arr->type->id()) {
     case arrow::Type::INT8:
-      return comp_view.AddPrimativeMemoryView<arrow::Int8Type>(arr);
+      return comp_view.AddPrimitiveMemoryView<arrow::Int8Type>(arr);
     case arrow::Type::INT16:
-      return comp_view.AddPrimativeMemoryView<arrow::Int16Type>(arr);
+      return comp_view.AddPrimitiveMemoryView<arrow::Int16Type>(arr);
     case arrow::Type::INT32:
-      return comp_view.AddPrimativeMemoryView<arrow::Int32Type>(arr);
+      return comp_view.AddPrimitiveMemoryView<arrow::Int32Type>(arr);
     case arrow::Type::INT64:
-      return comp_view.AddPrimativeMemoryView<arrow::Int64Type>(arr);
+      return comp_view.AddPrimitiveMemoryView<arrow::Int64Type>(arr);
     case arrow::Type::UINT8:
-      return comp_view.AddPrimativeMemoryView<arrow::UInt8Type>(arr);
+      return comp_view.AddPrimitiveMemoryView<arrow::UInt8Type>(arr);
     case arrow::Type::UINT16:
-      return comp_view.AddPrimativeMemoryView<arrow::UInt16Type>(arr);
+      return comp_view.AddPrimitiveMemoryView<arrow::UInt16Type>(arr);
     case arrow::Type::UINT32:
-      return comp_view.AddPrimativeMemoryView<arrow::UInt32Type>(arr);
+      return comp_view.AddPrimitiveMemoryView<arrow::UInt32Type>(arr);
     case arrow::Type::UINT64:
-      return comp_view.AddPrimativeMemoryView<arrow::UInt64Type>(arr);
+      return comp_view.AddPrimitiveMemoryView<arrow::UInt64Type>(arr);
     case arrow::Type::FLOAT:
-      return comp_view.AddPrimativeMemoryView<arrow::FloatType>(arr);
+      return comp_view.AddPrimitiveMemoryView<arrow::FloatType>(arr);
     case arrow::Type::DOUBLE:
-      return comp_view.AddPrimativeMemoryView<arrow::DoubleType>(arr);
+      return comp_view.AddPrimitiveMemoryView<arrow::DoubleType>(arr);
     case arrow::Type::DATE32:
-      return comp_view.AddPrimativeMemoryView<arrow::Date32Type>(arr);
+      return comp_view.AddPrimitiveMemoryView<arrow::Date32Type>(arr);
     case arrow::Type::DATE64:
-      return comp_view.AddPrimativeMemoryView<arrow::Date64Type>(arr);
+      return comp_view.AddPrimitiveMemoryView<arrow::Date64Type>(arr);
     case arrow::Type::TIME32:
-      return comp_view.AddPrimativeMemoryView<arrow::Time32Type>(arr);
+      return comp_view.AddPrimitiveMemoryView<arrow::Time32Type>(arr);
     case arrow::Type::TIME64:
-      return comp_view.AddPrimativeMemoryView<arrow::Time64Type>(arr);
+      return comp_view.AddPrimitiveMemoryView<arrow::Time64Type>(arr);
     case arrow::Type::TIMESTAMP:
-      return comp_view.AddPrimativeMemoryView<arrow::TimestampType>(arr);
+      return comp_view.AddPrimitiveMemoryView<arrow::TimestampType>(arr);
     case arrow::Type::DURATION:
-      return comp_view.AddPrimativeMemoryView<arrow::DurationType>(arr);
+      return comp_view.AddPrimitiveMemoryView<arrow::DurationType>(arr);
     case arrow::Type::INTERVAL_MONTHS:
-      return comp_view.AddPrimativeMemoryView<arrow::MonthIntervalType>(arr);
+      return comp_view.AddPrimitiveMemoryView<arrow::MonthIntervalType>(arr);
     // Need custom less function for this since it uses a custom struct for the data structure
     // case arrow::Type::INTERVAL_MONTH_DAY_NANO:
-    //   return comp_view.AddPrimativeMemoryView<arrow::MonthDayNanoIntervalType>(arr);
+    //   return comp_view.AddPrimitiveMemoryView<arrow::MonthDayNanoIntervalType>(arr);
     case arrow::Type::INTERVAL_DAY_TIME:
-      return comp_view.AddPrimativeMemoryView<arrow::DayTimeIntervalType>(arr);
+      return comp_view.AddPrimitiveMemoryView<arrow::DayTimeIntervalType>(arr);
     default:
       break;
   }
@@ -366,18 +372,18 @@ std::shared_ptr<arrow::Array> search_sorted_multiple_columns(const std::vector<s
   for (auto arr : sorted) {
     ARROW_CHECK_EQ(arr->GetNullCount(), 0);
     if (arrow::is_primitive(arr->type->id())) {
-      add_primative_memory_view_to_comp_view(sorted_comp_view, arr);
+      add_primitive_memory_view_to_comp_view(sorted_comp_view, arr);
     } else if (arrow::is_base_binary_like(arr->type->id())) {
       add_binary_memory_view_to_comp_view(sorted_comp_view, arr);
     } else {
       ARROW_LOG(FATAL) << "Unsupported Type " << arr->type->id();
     }
   }
-  bool key_has_nulls = false;
+  // bool key_has_nulls = false;
   for (auto arr : keys) {
-    key_has_nulls = key_has_nulls | (arr->GetNullCount() > 0);
+    // key_has_nulls = key_has_nulls | (arr->GetNullCount() > 0);
     if (arrow::is_primitive(arr->type->id())) {
-      add_primative_memory_view_to_comp_view(keys_comp_view, arr);
+      add_primitive_memory_view_to_comp_view(keys_comp_view, arr);
     } else if (arrow::is_base_binary_like(arr->type->id())) {
       add_binary_memory_view_to_comp_view(keys_comp_view, arr);
     } else {
@@ -392,26 +398,13 @@ std::shared_ptr<arrow::Array> search_sorted_multiple_columns(const std::vector<s
   size_t last_key_idx = 0;
 
   std::vector<std::shared_ptr<arrow::Buffer>> result_buffers{2};
-  if (key_has_nulls) {
-    result_buffers[0] = arrow::AllocateBitmap(key_len).ValueOrDie();
-  }
   result_buffers[1] = arrow::AllocateBuffer(sizeof(arrow::UInt64Type::c_type) * key_len).ValueOrDie();
   std::shared_ptr<arrow::ArrayData> result = arrow::ArrayData::Make(std::make_shared<arrow::UInt64Type>(), key_len, result_buffers);
 
   uint64_t *result_ptr = result->GetMutableValues<uint64_t>(1);
   ARROW_CHECK(result_ptr != NULL);
 
-  uint8_t *result_bitmask_ptr = result->GetMutableValues<uint8_t>(0);
-
   for (size_t key_idx = 0; key_idx < key_len; key_idx++, result_ptr++) {
-    if (key_has_nulls) {
-      const bool is_valid = keys_comp_view.isValid(key_idx);
-      bit_util::SetBitTo(result_bitmask_ptr, key_idx, is_valid);
-      if (!is_valid) {
-        continue;
-      }
-    }
-
     /*
      * Updating only one of the indices based on the previous key
      * gives the search a big boost when keys are sorted, but slightly
@@ -448,15 +441,11 @@ std::shared_ptr<arrow::Array> search_sorted_single_array(const arrow::Array *arr
   ARROW_CHECK(keys != NULL);
   const size_t size = keys->length();
   std::vector<std::shared_ptr<arrow::Buffer>> result_buffers{2};
-  if (keys->null_count() > 0) {
-    result_buffers[0] = arrow::AllocateBitmap(size).ValueOrDie();
-  }
   result_buffers[1] = arrow::AllocateBuffer(sizeof(arrow::UInt64Type::c_type) * size).ValueOrDie();
-  std::shared_ptr<arrow::ArrayData> result =
-      arrow::ArrayData::Make(std::make_shared<arrow::UInt64Type>(), size, result_buffers, keys->null_count());
+  std::shared_ptr<arrow::ArrayData> result = arrow::ArrayData::Make(std::make_shared<arrow::UInt64Type>(), size, result_buffers, 0);
   ARROW_CHECK(result.get() != NULL);
   if (arrow::is_primitive(arr->type()->id())) {
-    search_sorted_primative_single(arr->data().get(), keys->data().get(), result.get(), input_reversed);
+    search_sorted_primitive_single(arr->data().get(), keys->data().get(), result.get(), input_reversed);
   } else if (arrow::is_binary_like(arr->type()->id())) {
     search_sorted_binary_single<arrow::BinaryType::offset_type>(arr->data().get(), keys->data().get(), result.get(), input_reversed);
   } else if (arrow::is_large_binary_like(arr->type()->id())) {
