@@ -4,15 +4,18 @@ import itertools
 from abc import abstractmethod
 from dataclasses import dataclass
 from enum import Enum, IntEnum
-from typing import Any, Generic, List, Optional, Tuple, TypeVar, Union
+from pprint import pformat
+from typing import Any, Generic, TypeVar
 
 from daft.datasources import SourceInfo, StorageType
-from daft.expressions import ColumnExpression, Expression
+from daft.errors import ExpressionTypeError
+from daft.execution.operators import OperatorEnum
+from daft.expressions import CallExpression, ColumnExpression, Expression
 from daft.internal.treenode import TreeNode
 from daft.logical.map_partition_ops import ExplodeOp, MapPartitionOp
 from daft.logical.schema import ExpressionList
 from daft.resource_request import ResourceRequest
-from daft.runners.partitioning import vPartition
+from daft.runners.partitioning import PartitionCacheEntry, vPartition
 from daft.types import ExpressionType
 
 
@@ -87,11 +90,89 @@ class LogicalPlan(TreeNode["LogicalPlan"]):
     def rebuild(self) -> LogicalPlan:
         raise NotImplementedError()
 
+    @abstractmethod
+    def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
+        raise NotImplementedError()
+
+    def pretty_print(
+        self,
+    ) -> str:
+        builder: list[str] = []
+
+        def helper(node: LogicalPlan, depth: int = 0, index: int = 0, prefix: str = "", header: str = ""):
+            children: list[LogicalPlan] = node._children()
+            obj_repr_lines = repr(node).splitlines()
+            builder.append(f"{prefix}{header}{obj_repr_lines[0]}\n")
+
+            if len(children) > 0:
+                body_prefix = prefix + "│"
+            else:
+                body_prefix = prefix + " "
+
+            for line in obj_repr_lines[1:]:
+                builder.append(f"{body_prefix}{line}\n")
+            builder.append(f"{body_prefix}\n")
+
+            if len(children) < 2:
+                for child in children:
+                    has_grandchild = len(child._children()) > 0
+
+                    if has_grandchild:
+                        header = "├──"
+                    else:
+                        header = "└──"
+
+                    helper(child, depth=depth, index=index + 1, prefix=prefix, header=header)
+            else:
+
+                connector = "└─"
+                middle_child_header = "─┬─"
+
+                for i, child in enumerate(children):
+                    has_grandchild = len(child._children()) > 0
+                    if has_grandchild:
+                        final_header = "─┬─"
+                    else:
+                        final_header = "───"
+
+                    position = len(children) - i
+                    header = ("\b\b\b" * position) + connector + (middle_child_header * (position - 1)) + final_header
+                    if i != len(children) - 1:
+                        next_child_prefix = prefix + ("   │  " * (position - 1))
+                    else:
+                        next_child_prefix = prefix + "      "
+
+                    helper(child, depth=depth + 1, index=i, prefix=next_child_prefix, header=header)
+
+        helper(self, 0, 0, header="┌─")
+        return "".join(builder)
+
+    def _repr_helper(self, **fields: Any) -> str:
+
+        fields_to_print: dict[str, Any] = {}
+        if "output" not in fields:
+            fields_to_print["output"] = self.schema()
+
+        fields_to_print.update(fields)
+        fields_to_print["partitioning"] = self.partition_spec()
+        reduced_types = {}
+        for k, v in fields_to_print.items():
+            if isinstance(v, ExpressionList):
+                v = v.exprs
+            reduced_types[k] = v
+        to_render: list[str] = [f"{self.__class__.__name__}\n"]
+        space = "    "
+        for key, value in reduced_types.items():
+            repr_ed = pformat(value, width=80, compact=True).splitlines()
+            to_render.append(f"{space}{key}={repr_ed[0]}\n")
+            for line in repr_ed[1:]:
+                to_render.append(f"{space*2}{line}\n")
+
+        return "".join(to_render)
+
 
 class UnaryNode(LogicalPlan):
-    @abstractmethod
-    def copy_with_new_input(self, new_input: UnaryNode) -> UnaryNode:
-        raise NotImplementedError()
+    ...
 
 
 class BinaryNode(LogicalPlan):
@@ -104,8 +185,8 @@ class Scan(LogicalPlan):
         *,
         schema: ExpressionList,
         source_info: SourceInfo,
-        predicate: Optional[ExpressionList] = None,
-        columns: Optional[List[str]] = None,
+        predicate: ExpressionList | None = None,
+        columns: list[str] | None = None,
     ) -> None:
         schema = schema.resolve()
         pspec = PartitionSpec(scheme=PartitionScheme.UNKNOWN, num_partitions=source_info.get_num_partitions())
@@ -129,7 +210,7 @@ class Scan(LogicalPlan):
         return self._output_schema
 
     def __repr__(self) -> str:
-        return f"Scan\n\toutput={self.schema()}\n\tpredicate={self._predicate}\n\tcolumns={self._columns}\n\t{self._source_info}"
+        return self._repr_helper(columns_pruned=len(self._columns) - len(self.schema()), source_info=self._source_info)
 
     def resource_request(self) -> ResourceRequest:
         return ResourceRequest.default()
@@ -154,6 +235,50 @@ class Scan(LogicalPlan):
             columns=self._column_names,
         )
 
+    def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
+        assert len(new_children) == 0
+        return self
+
+
+class InMemoryScan(UnaryNode):
+    def __init__(
+        self, cache_entry: PartitionCacheEntry, schema: ExpressionList, partition_spec: PartitionSpec | None = None
+    ) -> None:
+
+        if partition_spec is None:
+            partition_spec = PartitionSpec(scheme=PartitionScheme.UNKNOWN, num_partitions=1)
+
+        super().__init__(schema=schema.resolve(), partition_spec=partition_spec, op_level=OpLevel.GLOBAL)
+        self._cache_entry = cache_entry
+
+    def __repr__(self) -> str:
+        return self._repr_helper(cache_id=self._cache_entry.key)
+
+    def _local_eq(self, other: Any) -> bool:
+        return (
+            isinstance(other, InMemoryScan)
+            and self._cache_entry == other._cache_entry
+            and self.schema() == other.schema()
+        )
+
+    def required_columns(self) -> ExpressionList:
+        return ExpressionList([])
+
+    def resource_request(self) -> ResourceRequest:
+        return ResourceRequest.default()
+
+    def rebuild(self) -> LogicalPlan:
+        # if we are rebuilding, this will be cached when this is ran
+        return InMemoryScan(
+            cache_entry=self._cache_entry,
+            schema=self.schema(),
+            partition_spec=self.partition_spec(),
+        )
+
+    def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
+        assert len(new_children) == 0
+        return self
+
 
 class FileWrite(UnaryNode):
     def __init__(
@@ -161,8 +286,8 @@ class FileWrite(UnaryNode):
         input: LogicalPlan,
         root_dir: str,
         storage_type: StorageType,
-        partition_cols: Optional[ExpressionList] = None,
-        compression: Optional[str] = None,
+        partition_cols: ExpressionList | None = None,
+        compression: str | None = None,
     ) -> None:
         assert (
             storage_type == StorageType.PARQUET or storage_type == StorageType.CSV
@@ -170,10 +295,10 @@ class FileWrite(UnaryNode):
         self._storage_type = storage_type
         self._root_dir = root_dir
         self._compression = compression
-        self._partition_cols = partition_cols
-        if self._partition_cols is not None:
-            self._partition_cols = self._partition_cols.resolve(input.schema())
-
+        if partition_cols is not None:
+            self._partition_cols = partition_cols.resolve(input.schema())
+        else:
+            self._partition_cols = ExpressionList([])
         for expr in input.schema():
             assert ExpressionType.is_primitive(
                 expr.resolved_type()
@@ -186,10 +311,10 @@ class FileWrite(UnaryNode):
         self._register_child(input)
 
     def __repr__(self) -> str:
-        return f"FileWrite\n\toutput={self.schema()}"
+        return self._repr_helper()
 
     def required_columns(self) -> ExpressionList:
-        raise NotImplementedError()
+        return self._partition_cols.required_columns()
 
     def resource_request(self) -> ResourceRequest:
         return ResourceRequest.default()
@@ -206,8 +331,15 @@ class FileWrite(UnaryNode):
     def rebuild(self) -> LogicalPlan:
         raise NotImplementedError("We can not rebuild a filewrite due to side effects")
 
-    def copy_with_new_input(self, new_input: UnaryNode) -> UnaryNode:
-        raise NotImplementedError()
+    def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
+        assert len(new_children) == 1
+        return FileWrite(
+            new_children[0],
+            root_dir=self._root_dir,
+            storage_type=self._storage_type,
+            partition_cols=self._partition_cols,
+            compression=self._compression,
+        )
 
 
 class Filter(UnaryNode):
@@ -227,7 +359,7 @@ class Filter(UnaryNode):
             )
 
     def __repr__(self) -> str:
-        return f"Filter\n\toutput={self.schema()}\n\tpredicate={self._predicate}"
+        return self._repr_helper(predicate=self._predicate)
 
     def resource_request(self) -> ResourceRequest:
         return self._predicate.resource_request()
@@ -238,12 +370,13 @@ class Filter(UnaryNode):
     def _local_eq(self, other: Any) -> bool:
         return isinstance(other, Filter) and self.schema() == other.schema() and self._predicate == other._predicate
 
-    def copy_with_new_input(self, new_input: LogicalPlan) -> Filter:
-        raise NotImplementedError()
-
     def rebuild(self) -> LogicalPlan:
         return Filter(input=self._children()[0].rebuild(), predicate=self._predicate.unresolve())
 
+    def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
+        assert len(new_children) == 1
+        return Filter(input=new_children[0], predicate=self._predicate)
+
 
 class Projection(UnaryNode):
     """Which columns to keep"""
@@ -255,7 +388,7 @@ class Projection(UnaryNode):
         self._projection = projection
 
     def __repr__(self) -> str:
-        return f"Projection\n\toutput={self.schema()}"
+        return self._repr_helper()
 
     def resource_request(self) -> ResourceRequest:
         return self._projection.resource_request()
@@ -268,35 +401,36 @@ class Projection(UnaryNode):
             isinstance(other, Projection) and self.schema() == other.schema() and self._projection == other._projection
         )
 
-    def copy_with_new_input(self, new_input: LogicalPlan) -> Projection:
-        return Projection(new_input, self._projection)
-
     def rebuild(self) -> LogicalPlan:
         return Projection(input=self._children()[0].rebuild(), projection=self._projection.unresolve())
 
+    def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
+        assert len(new_children) == 1
+        return Projection(new_children[0], self._projection)
+
 
 class Sort(UnaryNode):
-    def __init__(
-        self, input: LogicalPlan, sort_by: ExpressionList, descending: Union[List[bool], bool] = False
-    ) -> None:
+    def __init__(self, input: LogicalPlan, sort_by: ExpressionList, descending: list[bool] | bool = False) -> None:
         pspec = PartitionSpec(scheme=PartitionScheme.RANGE, num_partitions=input.num_partitions(), by=sort_by)
         super().__init__(input.schema().to_column_expressions(), partition_spec=pspec, op_level=OpLevel.GLOBAL)
         self._register_child(input)
         self._sort_by = sort_by.resolve(input_schema=input.schema())
+
+        for e in self._sort_by:
+            if e.resolved_type() == ExpressionType.null():
+                raise ExpressionTypeError(f"Cannot sort on null type expression: {e}")
+
         if isinstance(descending, bool):
             self._descending = [descending for _ in self._sort_by]
         else:
             self._descending = descending
 
     def __repr__(self) -> str:
-        return f"Sort\n\toutput={self.schema()}\n\tsort_by={self._sort_by}\n\tdesc={self._descending}"
+        return self._repr_helper(sort_by=self._sort_by, desc=self._descending)
 
     def resource_request(self) -> ResourceRequest:
         return self._sort_by.resource_request()
 
-    def copy_with_new_input(self, new_input: LogicalPlan) -> Sort:
-        return Sort(new_input, sort_by=self._sort_by, descending=self._descending)
-
     def required_columns(self) -> ExpressionList:
         return self._sort_by.required_columns()
 
@@ -311,6 +445,10 @@ class Sort(UnaryNode):
     def rebuild(self) -> LogicalPlan:
         return Sort(input=self._children()[0].rebuild(), sort_by=self._sort_by.unresolve(), descending=self._descending)
 
+    def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
+        assert len(new_children) == 1
+        return Sort(new_children[0], sort_by=self._sort_by, descending=self._descending)
+
 
 TMapPartitionOp = TypeVar("TMapPartitionOp", bound=MapPartitionOp)
 
@@ -326,7 +464,7 @@ class MapPartition(UnaryNode, Generic[TMapPartitionOp]):
         self._register_child(input)
 
     def __repr__(self) -> str:
-        return f"MapPartition\n\top={self._map_partition_op}"
+        return self._repr_helper(op=self._map_partition_op)
 
     def resource_request(self) -> ResourceRequest:
         return self._map_partition_op.resource_request()
@@ -344,15 +482,20 @@ class MapPartition(UnaryNode, Generic[TMapPartitionOp]):
 
 class Explode(MapPartition[ExplodeOp]):
     def __init__(self, input: LogicalPlan, explode_expressions: ExpressionList):
-        explode_expressions_resolved = ExpressionList([e._explode() for e in explode_expressions]).resolve(
-            input.schema()
-        )
+        assert [
+            isinstance(e, CallExpression) and e._operator == OperatorEnum.EXPLODE for e in explode_expressions
+        ], "Expressions supplied to Explode LogicalPlan must be a CallExpression with OperatorEnum.EXPLODE"
+
+        explode_expressions_resolved = explode_expressions.resolve(input.schema())
         map_partition_op = ExplodeOp(explode_columns=explode_expressions_resolved)
         super().__init__(
             input,
             map_partition_op,
         )
 
+    def __repr__(self) -> str:
+        return self._repr_helper()
+
     def required_columns(self) -> ExpressionList:
         return self._map_partition_op.explode_columns.required_columns()
 
@@ -362,8 +505,9 @@ class Explode(MapPartition[ExplodeOp]):
             self._map_partition_op.explode_columns.unresolve(),
         )
 
-    def copy_with_new_input(self, new_input: UnaryNode) -> Explode:
-        return Explode(new_input, explode_expressions=self._map_partition_op.explode_columns)
+    def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
+        assert len(new_children) == 1
+        return Explode(new_children[0], explode_expressions=self._map_partition_op.explode_columns)
 
 
 class LocalLimit(UnaryNode):
@@ -373,16 +517,17 @@ class LocalLimit(UnaryNode):
         self._num = num
 
     def __repr__(self) -> str:
-        return f"LocalLimit\n\toutput={self.schema()}\n\tN={self._num}"
+        return self._repr_helper(num=self._num)
 
     def resource_request(self) -> ResourceRequest:
         return ResourceRequest.default()
 
-    def copy_with_new_input(self, new_input: LogicalPlan) -> LocalLimit:
-        raise NotImplementedError()
+    def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
+        assert len(new_children) == 1
+        return LocalLimit(new_children[0], self._num)
 
     def required_columns(self) -> ExpressionList:
-        raise NotImplementedError()
+        return ExpressionList([])
 
     def _local_eq(self, other: Any) -> bool:
         return isinstance(other, LocalLimit) and self.schema() == other.schema() and self._num == self._num
@@ -398,16 +543,17 @@ class GlobalLimit(UnaryNode):
         self._num = num
 
     def __repr__(self) -> str:
-        return f"GlobalLimit\n\toutput={self.schema()}\n\tN={self._num}"
+        return self._repr_helper(num=self._num)
 
     def resource_request(self) -> ResourceRequest:
         return ResourceRequest.default()
 
-    def copy_with_new_input(self, new_input: LogicalPlan) -> GlobalLimit:
-        raise NotImplementedError()
+    def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
+        assert len(new_children) == 1
+        return GlobalLimit(new_children[0], self._num)
 
     def required_columns(self) -> ExpressionList:
-        raise NotImplementedError()
+        return ExpressionList([])
 
     def _local_eq(self, other: Any) -> bool:
         return isinstance(other, GlobalLimit) and self.schema() == other.schema() and self._num == self._num
@@ -422,12 +568,15 @@ class PartitionScheme(Enum):
     HASH = "HASH"
     RANDOM = "RANDOM"
 
+    def __repr__(self) -> str:
+        return self.value
+
 
 @dataclass(frozen=True)
 class PartitionSpec:
     scheme: PartitionScheme
     num_partitions: int
-    by: Optional[ExpressionList] = None
+    by: ExpressionList | None = None
 
 
 class Repartition(UnaryNode):
@@ -448,24 +597,24 @@ class Repartition(UnaryNode):
             raise ValueError("Can not pass in random partitioning and partition_by args")
 
     def __repr__(self) -> str:
-        return (
-            f"Repartition\n\toutput={self.schema()}\n\tpartition_by={self._partition_by}"
-            f"\n\tnum_partitions={self.num_partitions()}\n\tscheme={self._scheme}"
+        return self._repr_helper(
+            partition_by=self._partition_by, num_partitions=self.num_partitions(), scheme=self._scheme
         )
 
     def resource_request(self) -> ResourceRequest:
         return self._partition_by.resource_request()
 
-    def copy_with_new_input(self, new_input: LogicalPlan) -> Repartition:
+    def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
+        assert len(new_children) == 1
         return Repartition(
-            input=new_input,
+            input=new_children[0],
             partition_by=self._partition_by,
             num_partitions=self.num_partitions(),
             scheme=self._scheme,
         )
 
     def required_columns(self) -> ExpressionList:
-        raise NotImplementedError()
+        return self._partition_by.required_columns()
 
     def _local_eq(self, other: Any) -> bool:
         return (
@@ -498,19 +647,20 @@ class Coalesce(UnaryNode):
             )
 
     def __repr__(self) -> str:
-        return f"Coalesce\n\toutput={self.schema()}" f"\n\tnum_partitions={self.num_partitions()}"
+        return self._repr_helper(num_partitions=self.num_partitions())
 
     def resource_request(self) -> ResourceRequest:
         return ResourceRequest.default()
 
-    def copy_with_new_input(self, new_input: LogicalPlan) -> Coalesce:
+    def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
+        assert len(new_children) == 1
         return Coalesce(
-            input=new_input,
+            input=new_children[0],
             num_partitions=self.num_partitions(),
         )
 
     def required_columns(self) -> ExpressionList:
-        raise NotImplementedError()
+        return ExpressionList([])
 
     def _local_eq(self, other: Any) -> bool:
         return (
@@ -530,24 +680,24 @@ class LocalAggregate(UnaryNode):
     def __init__(
         self,
         input: LogicalPlan,
-        agg: List[Tuple[Expression, str]],
-        group_by: Optional[ExpressionList] = None,
+        agg: list[tuple[Expression, str]],
+        group_by: ExpressionList | None = None,
     ) -> None:
-
         cols_to_agg = ExpressionList([e for e, _ in agg]).resolve(input.schema())
         schema = cols_to_agg.to_column_expressions()
         self._group_by = group_by
-
+        self._required_cols = cols_to_agg.required_columns()
         if group_by is not None:
             self._group_by = group_by.resolve(input.schema())
             schema = self._group_by.union(schema)
+            self._required_cols = self._group_by.union(self._required_cols)
 
         super().__init__(schema, partition_spec=input.partition_spec(), op_level=OpLevel.PARTITION)
         self._register_child(input)
         self._agg = [(e, op) for e, (_, op) in zip(cols_to_agg, agg)]
 
     def __repr__(self) -> str:
-        return f"LocalAggregate\n\toutput={self.schema()}\n\tgroup_by={self._group_by}"
+        return self._repr_helper(agg=[e for e, _ in self._agg], group_by=self._group_by)
 
     def resource_request(self) -> ResourceRequest:
         req = ResourceRequest.default()
@@ -556,11 +706,12 @@ class LocalAggregate(UnaryNode):
         req = ResourceRequest.max_resources([expr.resource_request() for expr, _ in self._agg] + [req])
         return req
 
-    def copy_with_new_input(self, new_input: LogicalPlan) -> LocalAggregate:
-        raise NotImplementedError()
+    def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
+        assert len(new_children) == 1
+        return LocalAggregate(new_children[0], agg=self._agg, group_by=self._group_by)
 
     def required_columns(self) -> ExpressionList:
-        raise NotImplementedError()
+        return self._required_cols
 
     def _local_eq(self, other: Any) -> bool:
         return (
@@ -578,6 +729,44 @@ class LocalAggregate(UnaryNode):
         )
 
 
+class LocalDistinct(UnaryNode):
+    def __init__(
+        self,
+        input: LogicalPlan,
+        group_by: ExpressionList,
+    ) -> None:
+
+        self._group_by = group_by.resolve(input.schema())
+        schema = self._group_by.to_column_expressions()
+        super().__init__(schema, partition_spec=input.partition_spec(), op_level=OpLevel.PARTITION)
+        self._register_child(input)
+
+    def __repr__(self) -> str:
+        return self._repr_helper(group_by=self._group_by)
+
+    def resource_request(self) -> ResourceRequest:
+        req = ResourceRequest.default()
+        if self._group_by is not None:
+            req = self._group_by.resource_request()
+        req = ResourceRequest.max_resources([req])
+        return req
+
+    def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
+        assert len(new_children) == 1
+        return LocalDistinct(new_children[0], group_by=self._group_by)
+
+    def required_columns(self) -> ExpressionList:
+        return self._group_by.required_columns()
+
+    def _local_eq(self, other: Any) -> bool:
+        return (
+            isinstance(other, LocalDistinct) and self.schema() == other.schema() and self._group_by == other._group_by
+        )
+
+    def rebuild(self) -> LogicalPlan:
+        return LocalDistinct(input=self._children()[0].rebuild(), group_by=self._group_by.unresolve())
+
+
 class HTTPRequest(LogicalPlan):
     def __init__(
         self,
@@ -591,7 +780,7 @@ class HTTPRequest(LogicalPlan):
         return self._output_schema
 
     def __repr__(self) -> str:
-        return f"HTTPRequest\n\toutput={self.schema()}"
+        return self._repr_helper()
 
     def resource_request(self) -> ResourceRequest:
         return ResourceRequest.default()
@@ -605,6 +794,10 @@ class HTTPRequest(LogicalPlan):
     def rebuild(self) -> LogicalPlan:
         return HTTPRequest(schema=self.schema().unresolve())
 
+    def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
+        assert len(new_children) == 0
+        return self
+
 
 class HTTPResponse(UnaryNode):
     def __init__(
@@ -618,7 +811,7 @@ class HTTPResponse(UnaryNode):
         return self._schema
 
     def __repr__(self) -> str:
-        return f"HTTPResponse\n\toutput={self.schema()}"
+        return self._repr_helper()
 
     def resource_request(self) -> ResourceRequest:
         return ResourceRequest.default()
@@ -629,8 +822,9 @@ class HTTPResponse(UnaryNode):
     def _local_eq(self, other: Any) -> bool:
         return isinstance(other, HTTPResponse) and self.schema() == other.schema()
 
-    def copy_with_new_input(self, new_input: LogicalPlan) -> HTTPResponse:
-        raise NotImplementedError()
+    def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
+        assert len(new_children) == 1
+        return HTTPResponse(new_children[0])
 
     def rebuild(self) -> LogicalPlan:
         return HTTPResponse(
@@ -661,6 +855,12 @@ class Join(BinaryNode):
         num_partitions: int
         self._left_on = left_on.resolve(left.schema())
         self._right_on = right_on.resolve(right.schema())
+
+        for join_keys in (self._left_on, self._right_on):
+            for e in join_keys:
+                if e.resolved_type() == ExpressionType.null():
+                    raise ExpressionTypeError(f"Cannot join on null type expression: {e}")
+
         self._how = how
         schema: ExpressionList
         if how == JoinType.LEFT:
@@ -670,15 +870,30 @@ class Join(BinaryNode):
             num_partitions = right.num_partitions()
             raise NotImplementedError()
         elif how == JoinType.INNER:
-            num_partitions = min(left.num_partitions(), right.num_partitions())
+            num_partitions = max(left.num_partitions(), right.num_partitions())
             right_id_set = self._right_on.to_id_set()
             filtered_right = [e for e in right.schema() if e.get_id() not in right_id_set]
             schema = left.schema().union(ExpressionList(filtered_right), rename_dup="right.")
 
-        left = Repartition(left, partition_by=self._left_on, num_partitions=num_partitions, scheme=PartitionScheme.HASH)
-        right = Repartition(
+        left_pspec = PartitionSpec(scheme=PartitionScheme.HASH, num_partitions=num_partitions, by=self._left_on)
+        right_pspec = PartitionSpec(scheme=PartitionScheme.HASH, num_partitions=num_partitions, by=self._right_on)
+
+        new_left = Repartition(
+            left, partition_by=self._left_on, num_partitions=num_partitions, scheme=PartitionScheme.HASH
+        )
+
+        if num_partitions == 1 and left.num_partitions() == 1:
+            left = left
+        elif left.partition_spec() != left_pspec:
+            left = new_left
+
+        new_right = Repartition(
             right, partition_by=self._right_on, num_partitions=num_partitions, scheme=PartitionScheme.HASH
         )
+        if num_partitions == 1 and right.num_partitions() == 1:
+            right = right
+        elif right.partition_spec() != right_pspec:
+            right = new_right
 
         super().__init__(
             schema.to_column_expressions(), partition_spec=left.partition_spec(), op_level=OpLevel.PARTITION
@@ -687,12 +902,7 @@ class Join(BinaryNode):
         self._register_child(right)
 
     def __repr__(self) -> str:
-        return (
-            f"Join\n\toutput={self.schema()}"
-            f"\n\tnum_partitions={self.num_partitions()}"
-            f"\n\tleft_on={self._left_on}"
-            f"\n\tright_on={self._right_on}"
-        )
+        return self._repr_helper(left_on=self._left_on, right_on=self._right_on, num_partitions=self.num_partitions())
 
     def resource_request(self) -> ResourceRequest:
         # Note that this join creates two Repartition LogicalPlans using the left_on and right_on ExpressionLists
@@ -700,11 +910,12 @@ class Join(BinaryNode):
         # after repartitioning is done should be relatively cheap.
         return ResourceRequest.default()
 
-    def copy_with_new_input(self, new_input: LogicalPlan) -> Coalesce:
-        raise NotImplementedError()
+    def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
+        assert len(new_children) == 2
+        return Join(new_children[0], new_children[1], left_on=self._left_on, right_on=self._right_on, how=self._how)
 
     def required_columns(self) -> ExpressionList:
-        raise NotImplementedError()
+        return self._left_on.required_columns().union(self._right_on.required_columns(), rename_dup="right.")
 
     def _local_eq(self, other: Any) -> bool:
         return (
