@@ -1,6 +1,8 @@
 from __future__ import annotations
 
 import collections
+import datetime
+import functools
 import math
 import operator
 import random
@@ -14,10 +16,7 @@ from typing import (
     Iterable,
     Iterator,
     List,
-    Optional,
     Sequence,
-    Tuple,
-    Type,
     TypeVar,
     Union,
     cast,
@@ -30,8 +29,9 @@ import pyarrow.compute as pac
 from pandas.core.reshape.merge import get_join_indexers
 
 from daft.execution.operators import OperatorEnum, OperatorEvaluator
-from daft.internal.hashing import hash_chunked_array
+from daft.internal.kernels.hashing import hash_chunked_array
 from daft.internal.kernels.search_sorted import search_sorted
+from daft.types import ExpressionType, PrimitiveExpressionType, PythonExpressionType
 
 # A type representing some Python scalar (non-series/array like object)
 PyScalar = Any
@@ -43,7 +43,7 @@ BinaryFuncType = Callable[[ArrType, ArrType], ArrType]
 NUMPY_MINOR_VERSION = int(np.version.version.split(".")[1])
 
 
-def zip_blocks_as_py(*blocks: DataBlock) -> Iterator[Tuple[Any, ...]]:
+def zip_blocks_as_py(*blocks: DataBlock) -> Iterator[tuple[Any, ...]]:
     """Utility to zip the data of blocks together, returning a row-based iterator. This utility
     accounts for special-cases such as blocks that contain a single value to be broadcasted,
     differences in the underlying data representation etc.
@@ -76,7 +76,7 @@ def zip_blocks_as_py(*blocks: DataBlock) -> Iterator[Tuple[Any, ...]]:
 
 class DataBlock(Generic[ArrType]):
     data: ArrType
-    evaluator: ClassVar[Type[OperatorEvaluator]]
+    evaluator: ClassVar[type[OperatorEvaluator]]
 
     def __init__(self, data: ArrType) -> None:
         self.data = data
@@ -110,7 +110,7 @@ class DataBlock(Generic[ArrType]):
         raise NotImplementedError()
 
     @abstractmethod
-    def to_polars(self) -> Union[PyScalar, pl.Series]:
+    def to_polars(self) -> PyScalar | pl.Series:
         raise NotImplementedError()
 
     @abstractmethod
@@ -197,7 +197,7 @@ class DataBlock(Generic[ArrType]):
 
     def partition(
         self, num: int, pivots: np.ndarray, target_partitions: np.ndarray, argsorted_targets: DataBlock[ArrowArrType]
-    ) -> List[DataBlock[ArrType]]:
+    ) -> list[DataBlock[ArrType]]:
         assert not self.is_scalar(), "Cannot partition scalar DataBlock"
 
         # We first argsort the targets to group the same partitions together
@@ -256,11 +256,11 @@ class DataBlock(Generic[ArrType]):
 
     @staticmethod
     @abstractmethod
-    def _argsort(blocks: List[DataBlock[ArrType]], descending: Optional[List[bool]] = None) -> DataBlock[ArrowArrType]:
+    def _argsort(blocks: list[DataBlock[ArrType]], descending: list[bool] | None = None) -> DataBlock[ArrowArrType]:
         raise NotImplementedError()
 
     @classmethod
-    def argsort(cls, blocks: List[DataBlock], descending: Optional[List[bool]] = None) -> DataBlock[ArrowArrType]:
+    def argsort(cls, blocks: list[DataBlock], descending: list[bool] | None = None) -> DataBlock[ArrowArrType]:
         first_type = type(blocks[0])
         assert all(type(b) == first_type for b in blocks), "all block types must match"
         size = len(blocks)
@@ -273,13 +273,13 @@ class DataBlock(Generic[ArrType]):
     @staticmethod
     @abstractmethod
     def _search_sorted(
-        sorted_blocks: List[DataBlock], keys: List[DataBlock], input_reversed: Optional[List[bool]] = None
+        sorted_blocks: list[DataBlock], keys: list[DataBlock], input_reversed: list[bool] | None = None
     ) -> DataBlock[ArrowArrType]:
         raise NotImplementedError()
 
     @classmethod
     def search_sorted(
-        cls, sorted_blocks: List[DataBlock], keys: List[DataBlock], input_reversed: Optional[List[bool]] = None
+        cls, sorted_blocks: list[DataBlock], keys: list[DataBlock], input_reversed: list[bool] | None = None
     ) -> DataBlock[ArrowArrType]:
         first_type = type(sorted_blocks[0])
         assert all(type(b) == first_type for b in sorted_blocks), "all block types must match"
@@ -295,18 +295,18 @@ class DataBlock(Generic[ArrType]):
 
     @staticmethod
     @abstractmethod
-    def _merge_blocks(blocks: List[DataBlock[ArrType]]) -> DataBlock[ArrType]:
+    def _merge_blocks(blocks: list[DataBlock[ArrType]]) -> DataBlock[ArrType]:
         raise NotImplementedError()
 
     @classmethod
-    def merge_blocks(cls, blocks: List[DataBlock[ArrType]]) -> DataBlock[ArrType]:
+    def merge_blocks(cls, blocks: list[DataBlock[ArrType]]) -> DataBlock[ArrType]:
         assert len(blocks) > 0, "no blocks"
         first_type = type(blocks[0])
         assert all(type(b) == first_type for b in blocks), "all block types must match"
         return first_type._merge_blocks(blocks)
 
     @abstractmethod
-    def array_hash(self, seed: Optional[DataBlock[ArrType]] = None) -> DataBlock[ArrowArrType]:
+    def array_hash(self, seed: DataBlock[ArrType] | None = None) -> DataBlock[ArrowArrType]:
         raise NotImplementedError()
 
     @abstractmethod
@@ -316,20 +316,23 @@ class DataBlock(Generic[ArrType]):
     @staticmethod
     @abstractmethod
     def _group_by_agg(
-        group_by: List[DataBlock[ArrType]], to_agg: List[DataBlock[ArrType]], agg_ops: List[str]
-    ) -> Tuple[List[DataBlock[ArrType]], List[DataBlock[ArrType]]]:
+        group_by: list[DataBlock[ArrType]], to_agg: list[DataBlock[ArrType]], agg_ops: list[str]
+    ) -> tuple[list[DataBlock[ArrType]], list[DataBlock[ArrType]]]:
         raise NotImplementedError()
 
     @classmethod
     def group_by_agg(
-        cls, group_by: List[DataBlock[ArrType]], to_agg: List[DataBlock[ArrType]], agg_ops: List[str]
-    ) -> Tuple[List[DataBlock[ArrType]], List[DataBlock[ArrType]]]:
+        cls, group_by: list[DataBlock[ArrType]], to_agg: list[DataBlock[ArrType]], agg_ops: list[str]
+    ) -> tuple[list[DataBlock[ArrType]], list[DataBlock[ArrType]]]:
         assert len(group_by) > 0, "no blocks"
-        assert len(to_agg) > 0, "no blocks"
         assert len(to_agg) == len(agg_ops)
-        first_type = type(to_agg[0])
+        if len(to_agg) > 0:
+            first_type = type(to_agg[0])
+            assert all(type(b) == first_type for b in to_agg), "all block types must match"
+        else:
+            first_type = type(group_by[0])
+
         assert all(type(b) == first_type for b in group_by), "all block types must match"
-        assert all(type(b) == first_type for b in to_agg), "all block types must match"
 
         return first_type._group_by_agg(group_by, to_agg, agg_ops)
 
@@ -339,18 +342,18 @@ class DataBlock(Generic[ArrType]):
     @staticmethod
     @abstractmethod
     def _join_keys(
-        left_keys: List[DataBlock[ArrType]], right_keys: List[DataBlock[ArrType]]
-    ) -> Tuple[DataBlock[ArrType], DataBlock[ArrType]]:
+        left_keys: list[DataBlock[ArrType]], right_keys: list[DataBlock[ArrType]]
+    ) -> tuple[DataBlock[ArrType], DataBlock[ArrType]]:
         raise NotImplementedError()
 
     @classmethod
     def join(
         cls,
-        left_keys: List[DataBlock],
-        right_keys: List[DataBlock],
-        left_columns: List[DataBlock],
-        right_columns: List[DataBlock],
-    ) -> List[DataBlock]:
+        left_keys: list[DataBlock],
+        right_keys: list[DataBlock],
+        left_columns: list[DataBlock],
+        right_columns: list[DataBlock],
+    ) -> list[DataBlock]:
         assert len(left_keys) > 0
         assert len(left_keys) == len(right_keys)
         last_type = None
@@ -368,11 +371,13 @@ class DataBlock(Generic[ArrType]):
 
         for b in right_columns:
             to_rtn.append(b.take(right_indices))
-
+        all_lens = set(map(len, to_rtn))
+        if len(all_lens) != 1:
+            raise ValueError(f"expected all lengths to be the same, got {list(map(len, to_rtn))}")
         return to_rtn
 
     @abstractmethod
-    def list_explode(self) -> Tuple[DataBlock[ArrType], DataBlock[ArrowArrType]]:
+    def list_explode(self) -> tuple[DataBlock[ArrType], DataBlock[ArrowArrType]]:
         """Treats each row as a list and flattens the nested lists. Will throw an error if
         elements in the block cannot be treated as a list.
 
@@ -407,7 +412,7 @@ class PyListDataBlock(DataBlock[List[T]]):
             res[i] = obj
         return res
 
-    def to_polars(self) -> Union[PyScalar, pl.Series]:
+    def to_polars(self) -> PyScalar | pl.Series:
         if self.is_scalar():
             return self.data
         return pl.Series(self.data, dtype=pl.datatypes.Object)
@@ -415,39 +420,39 @@ class PyListDataBlock(DataBlock[List[T]]):
     def to_arrow(self) -> pa.ChunkedArray:
         raise NotImplementedError("can not convert pylist block to arrow")
 
-    def _filter(self, mask: DataBlock[ArrowArrType]) -> DataBlock[List[T]]:
+    def _filter(self, mask: DataBlock[ArrowArrType]) -> DataBlock[list[T]]:
         return PyListDataBlock(data=[item for keep, item in zip(mask.iter_py(), self.data) if keep])
 
-    def _take(self, indices: DataBlock[ArrowArrType]) -> DataBlock[List[T]]:
+    def _take(self, indices: DataBlock[ArrowArrType]) -> DataBlock[list[T]]:
         return PyListDataBlock(data=[self.data[i] for i in indices.iter_py()])
 
-    def sample(self, num: int) -> DataBlock[List[T]]:
+    def sample(self, num: int) -> DataBlock[list[T]]:
         return PyListDataBlock(data=random.sample(self.data, num))
 
     @staticmethod
     def _search_sorted(
-        sorted_blocks: List[DataBlock], keys: List[DataBlock], input_reversed: Optional[List[bool]] = None
+        sorted_blocks: list[DataBlock], keys: list[DataBlock], input_reversed: list[bool] | None = None
     ) -> DataBlock[ArrowArrType]:
         raise NotImplementedError()
 
     @staticmethod
-    def _argsort(blocks: List[DataBlock[ArrType]], descending: Optional[List[bool]] = None) -> DataBlock[ArrowArrType]:
+    def _argsort(blocks: list[DataBlock[ArrType]], descending: list[bool] | None = None) -> DataBlock[ArrowArrType]:
         raise NotImplementedError("Sorting by Python objects is not implemented")
 
-    def _make_empty(self) -> DataBlock[List[T]]:
+    def _make_empty(self) -> DataBlock[list[T]]:
         return PyListDataBlock(data=[])
 
-    def _split(self, pivots: np.ndarray) -> Sequence[List[T]]:
+    def _split(self, pivots: np.ndarray) -> Sequence[list[T]]:
         return [list(chunk) for chunk in np.split(self.data, pivots)[1:]]
 
     @staticmethod
-    def _merge_blocks(blocks: List[DataBlock[List[T]]]) -> DataBlock[List[T]]:
+    def _merge_blocks(blocks: list[DataBlock[list[T]]]) -> DataBlock[list[T]]:
         concatted_data = []
         for block in blocks:
             concatted_data.extend(block.data)
         return PyListDataBlock(data=concatted_data)
 
-    def array_hash(self, seed: Optional[DataBlock[ArrType]] = None) -> DataBlock[ArrowArrType]:
+    def array_hash(self, seed: DataBlock[ArrType] | None = None) -> DataBlock[ArrowArrType]:
         # TODO(jay): seed is ignored here, but perhaps we need to set it in PYTHONHASHSEED?
         hashes = [hash(x) for x in self.data]
         return DataBlock.make_block(np.array(hashes))
@@ -457,17 +462,17 @@ class PyListDataBlock(DataBlock[List[T]]):
 
     @staticmethod
     def _group_by_agg(
-        group_by: List[DataBlock[List[T]]], to_agg: List[DataBlock[List[T]]], agg_ops: List[str]
-    ) -> Tuple[List[DataBlock[List[T]]], List[DataBlock[List[T]]]]:
+        group_by: list[DataBlock[list[T]]], to_agg: list[DataBlock[list[T]]], agg_ops: list[str]
+    ) -> tuple[list[DataBlock[list[T]]], list[DataBlock[list[T]]]]:
         raise NotImplementedError("Aggregations on Python objects is not implemented yet")
 
     @staticmethod
     def _join_keys(
-        left_keys: List[DataBlock[List[T]]], right_keys: List[DataBlock[List[T]]]
-    ) -> Tuple[DataBlock[List[T]], DataBlock[List[T]]]:
+        left_keys: list[DataBlock[list[T]]], right_keys: list[DataBlock[list[T]]]
+    ) -> tuple[DataBlock[list[T]], DataBlock[list[T]]]:
         raise NotImplementedError()
 
-    def list_explode(self) -> Tuple[PyListDataBlock, DataBlock[ArrowArrType]]:
+    def list_explode(self) -> tuple[PyListDataBlock, DataBlock[ArrowArrType]]:
         lengths = []
         exploded = []
         for item in self.data:
@@ -498,7 +503,7 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
         ), f"Cannot create ArrowDataBlock with nested type: {data.type}, please report this to Daft developers."
         super().__init__(data)
 
-    def __reduce__(self) -> Tuple:
+    def __reduce__(self) -> tuple:
         if len(self.data) == 0:
             return ArrowDataBlock, (self._make_empty().data,)
         else:
@@ -509,7 +514,7 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
             return self.data.as_py()
         return self.data.to_numpy()
 
-    def to_polars(self) -> Union[PyScalar, pl.Series]:
+    def to_polars(self) -> PyScalar | pl.Series:
         if isinstance(self.data, pa.Scalar):
             return self.data.as_py()
         return pl.Series(self.data)
@@ -533,7 +538,7 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
     #     return ArrowDataBlock(data=sort_indices)
 
     @staticmethod
-    def _argsort(blocks: List[DataBlock[ArrType]], descending: Optional[List[bool]] = None) -> DataBlock[ArrowArrType]:
+    def _argsort(blocks: list[DataBlock[ArrType]], descending: list[bool] | None = None) -> DataBlock[ArrowArrType]:
         arrs = [a.data for a in blocks]
         arr_names = [f"a_{i}" for i in range(len(blocks))]
         if descending is None:
@@ -545,9 +550,8 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
 
     @staticmethod
     def _search_sorted(
-        sorted_blocks: List[DataBlock], keys: List[DataBlock], input_reversed: Optional[List[bool]] = None
+        sorted_blocks: list[DataBlock], keys: list[DataBlock], input_reversed: list[bool] | None = None
     ) -> DataBlock[ArrowArrType]:
-
         arr_names = [f"a_{i}" for i in range(len(sorted_blocks))]
         if input_reversed is None:
             input_reversed = [False for _ in range(len(sorted_blocks))]
@@ -573,9 +577,16 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
                 num_chunks = math.ceil(expected_size / (2**31))
                 chunk_size = len(indices) // num_chunks
                 new_indices_subarray = []
-                for i in range(num_chunks):
+                for i in range(num_chunks - 1):
                     new_indices_subarray.extend(indices.data.slice(i * chunk_size, chunk_size).chunks)
+
+                new_indices_subarray.extend(indices.data.slice((num_chunks - 1) * chunk_size).chunks)
+                old_indices = indices
                 indices = DataBlock.make_block(pa.chunked_array(new_indices_subarray))
+                if len(old_indices) != len(indices):
+                    raise ValueError(
+                        f"expected new indices to be the same length as the old one {len(old_indices)} bs {len(indices)}"
+                    )
 
         return self._binary_op(indices, fn=partial(pac.take, boundscheck=False))
 
@@ -591,11 +602,13 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
         return to_return
 
     @staticmethod
-    def _merge_blocks(blocks: List[DataBlock[ArrowArrType]]) -> DataBlock[ArrowArrType]:
+    def _merge_blocks(blocks: list[DataBlock[ArrowArrType]]) -> DataBlock[ArrowArrType]:
+        block_types = {block.data.type for block in blocks}
+        assert len(block_types) == 1, f"Unable to merge blocks of different types: {block_types}"
         all_chunks = []
         for block in blocks:
             all_chunks.extend(block.data.chunks)
-        return DataBlock.make_block(pa.chunked_array(all_chunks))
+        return DataBlock.make_block(pa.chunked_array(all_chunks, type=block_types.pop()))
 
     def _make_empty(self) -> DataBlock[ArrowArrType]:
         return DataBlock.make_block(data=pa.chunked_array([[]], type=self.data.type))
@@ -606,54 +619,41 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
         sampled = np.random.choice(self.data, num, replace=False)
         return DataBlock.make_block(data=pa.chunked_array([sampled]))
 
-    def array_hash(self, seed: Optional[DataBlock[ArrowArrType]] = None) -> DataBlock[ArrowArrType]:
+    def array_hash(self, seed: DataBlock[ArrowArrType] | None = None) -> DataBlock[ArrowArrType]:
         assert isinstance(self.data, pa.ChunkedArray)
         assert seed is None or isinstance(seed.data, pa.ChunkedArray)
 
-        pa_type = self.data.type
-        data_to_hash = self.data
-        if pa.types.is_date32(pa_type):
-            data_to_hash = data_to_hash.cast(pa.int32())
-        elif pa.types.is_date64(pa_type):
-            data_to_hash = data_to_hash.cast(pa.int64())
-        elif pa.types.is_floating(pa_type):
-            data_to_hash = data_to_hash.cast(pa.string())
-        elif not (pa.types.is_integer(pa_type) or pa.types.is_string(pa_type)):
-            raise TypeError(f"cannot hash {pa_type}")
-        hashed = hash_chunked_array(data_to_hash)
-        if seed is None:
-            return DataBlock.make_block(data=hashed)
-        else:
-            seed_pa_type = seed.data.type
-            if not (pa.types.is_uint64(seed_pa_type)):
-                raise TypeError(f"can only seed hash uint64 not {seed_pa_type}")
-
-            seed_arr = seed.data.to_numpy()
-            seed_arr = seed_arr ^ (hashed.to_numpy() + 0x9E3779B9 + (seed_arr << 6) + (seed_arr >> 2))
-            return DataBlock.make_block(data=pa.chunked_array([seed_arr]))
+        hashed = hash_chunked_array(self.data, seed.data if seed is not None else None)
+        return DataBlock.make_block(data=hashed)
 
     def agg(self, op: str) -> DataBlock[ArrowArrType]:
-
         if op == "sum":
             if len(self) == 0:
                 return ArrowDataBlock(data=pa.chunked_array([[]], type=self.data.type))
-            return ArrowDataBlock(data=pa.chunked_array([[pac.sum(self.data).as_py()]]))
+            return ArrowDataBlock(data=pa.chunked_array([[pac.sum(self.data, min_count=0).as_py()]]))
         elif op == "mean":
             if len(self) == 0:
                 return ArrowDataBlock(data=pa.chunked_array([[]], type=pa.float64()))
-            return ArrowDataBlock(data=pa.chunked_array([[pac.mean(self.data).as_py()]]))
+            return ArrowDataBlock(data=pa.chunked_array([[pac.mean(self.data).as_py()]], type=pa.float64()))
         elif op == "count":
             if len(self) == 0:
                 return ArrowDataBlock(data=pa.chunked_array([[]], type=pa.int64()))
-            return ArrowDataBlock(data=pa.chunked_array([[pac.count(self.data).as_py()]]))
-
+            return ArrowDataBlock(data=pa.chunked_array([[pac.count(self.data).as_py()]], type=pa.int64()))
+        elif op == "min":
+            if len(self) == 0:
+                return ArrowDataBlock(data=pa.chunked_array([[]], type=self.data.type))
+            return ArrowDataBlock(data=pa.chunked_array([[pac.min(self.data).as_py()]], type=self.data.type))
+        elif op == "max":
+            if len(self) == 0:
+                return ArrowDataBlock(data=pa.chunked_array([[]], type=self.data.type))
+            return ArrowDataBlock(data=pa.chunked_array([[pac.max(self.data).as_py()]], type=self.data.type))
         else:
             raise NotImplementedError(op)
 
     @staticmethod
     def _group_by_agg(
-        group_by: List[DataBlock[ArrowArrType]], to_agg: List[DataBlock[ArrowArrType]], agg_ops: List[str]
-    ) -> Tuple[List[DataBlock[ArrowArrType]], List[DataBlock[ArrowArrType]]]:
+        group_by: list[DataBlock[ArrowArrType]], to_agg: list[DataBlock[ArrowArrType]], agg_ops: list[str]
+    ) -> tuple[list[DataBlock[ArrowArrType]], list[DataBlock[ArrowArrType]]]:
         group_arrs = [a.data for a in group_by]
         agg_arrs = [a.data for a in to_agg]
         arrs = group_arrs + agg_arrs
@@ -679,35 +679,73 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
                 exprs.append(pl.max(an))
                 agg_expected_arrow_type.append(arr.type)
             elif op == "count":
-                exprs.append(pl.count(an))
+                exprs.append(pl.col(an).is_not_null().sum().alias(an))
                 agg_expected_arrow_type.append(pa.int64())
             else:
                 raise NotImplementedError()
+        if len(agg_names) == 0:
+            exprs.append(pl.count(group_names[0]).alias("placeholder"))
+            agg_expected_arrow_type.append(pa.int64())
 
         pl_agged = pl_table.groupby(group_names).agg(exprs)
         agged = pl_agged.to_arrow()
-        gcols: List[DataBlock] = [
+        gcols: list[DataBlock] = [
             DataBlock.make_block(agged[g_name].cast(t)) for g_name, t in zip(group_names, grouped_expected_arrow_type)
         ]
-        acols: List[DataBlock] = [
+        acols: list[DataBlock] = [
             DataBlock.make_block(agged[f"{a_name}"].cast(t)) for a_name, t in zip(agg_names, agg_expected_arrow_type)
         ]
+
+        if len(agg_names) == 0:
+            acols = []
+
         return gcols, acols
 
+    @staticmethod
+    def _get_null_nan_mask(keys: list[DataBlock[ArrowArrType]]) -> pa.Array:
+        """Returns a mask which indicates the presence of nulls/nans across a list of DataBlocks"""
+        mask_list = [pac.invert(pac.is_null(k.data, nan_is_null=True)) for k in keys]
+        mask = functools.reduce(lambda a, b: pac.and_(a, b), mask_list)
+        return mask
+
     @staticmethod
     def _join_keys(
-        left_keys: List[DataBlock[ArrowArrType]], right_keys: List[DataBlock[ArrowArrType]]
-    ) -> Tuple[DataBlock[ArrowArrType], DataBlock[ArrowArrType]]:
+        left_keys: list[DataBlock[ArrowArrType]], right_keys: list[DataBlock[ArrowArrType]]
+    ) -> tuple[DataBlock[ArrowArrType], DataBlock[ArrowArrType]]:
         assert len(left_keys) == len(right_keys)
+        left_null_nan_mask = ArrowDataBlock._get_null_nan_mask(left_keys)
+        left_no_nulls = pac.all(left_null_nan_mask).as_py()
+        right_null_nan_mask = ArrowDataBlock._get_null_nan_mask(right_keys)
+        right_no_nulls = pac.all(right_null_nan_mask).as_py()
 
-        pd_left_keys = [k.data.to_pandas() for k in left_keys]
+        # Filter out rows with NaNs/None entries
+        left_keys_arrs = (
+            [k.data for k in left_keys]
+            if left_no_nulls
+            else [pac.array_filter(k.data, left_null_nan_mask) for k in left_keys]
+        )
+        right_keys_arrs = (
+            [k.data for k in right_keys]
+            if right_no_nulls
+            else [pac.array_filter(k.data, right_null_nan_mask) for k in right_keys]
+        )
 
-        pd_right_keys = [k.data.to_pandas() for k in right_keys]
-        left_index, right_index = get_join_indexers(pd_left_keys, pd_right_keys, how="inner")
+        left_index, right_index = get_join_indexers(
+            [arr.to_pandas() for arr in left_keys_arrs], [arr.to_pandas() for arr in right_keys_arrs], how="inner"
+        )
 
+        # Restore the join indices to original indices before filtering rows with NaNs/Nones
+        if not left_no_nulls:
+            num_nulls_cumsum = np.invert(left_null_nan_mask.to_numpy()).cumsum()[left_null_nan_mask.to_numpy()]
+            left_index = left_index + num_nulls_cumsum[left_index]
+        if not right_no_nulls:
+            num_nulls_cumsum = np.invert(right_null_nan_mask.to_numpy()).cumsum()[right_null_nan_mask.to_numpy()]
+            right_index = right_index + num_nulls_cumsum[right_index]
+
+        # NOTE: the results presented here are always an INNER join because all the NaN/Null keys are always filtered out
         return DataBlock.make_block(left_index), DataBlock.make_block(right_index)
 
-    def list_explode(self) -> Tuple[DataBlock[ArrType], DataBlock[ArrowArrType]]:
+    def list_explode(self) -> tuple[DataBlock[ArrType], DataBlock[ArrowArrType]]:
         """Treats each row as a list and flattens the nested lists. Will throw an error if
         elements in the block cannot be treated as a list.
         """
@@ -723,9 +761,11 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
 
 
 def arrow_mod(arr, m):
-    if isinstance(m, pa.Scalar):
-        return np.mod(arr, m.as_py())
-    return np.mod(arr, m)
+    # TODO: Write a better kernel for mod without relying on Numpy, but respects typing semantics of Polars (see: https://github.com/Eventual-Inc/Daft/issues/250)
+    mask = arr.is_null()
+    modded_np = np.mod(arr, m.as_py()) if isinstance(m, pa.Scalar) else np.mod(arr, m)
+    return_type = pa.int64() if pa.types.is_integer(m.type) and pa.types.is_integer(arr.type) else pa.float64()
+    return pa.array(modded_np, mask=mask.to_numpy(), type=return_type)
 
 
 def arrow_floordiv(arr, m):
@@ -754,6 +794,24 @@ def arrow_str_startswith(arr: pa.ChunkedArray, pattern: pa.StringScalar):
     return pac.starts_with(arr, pattern=pattern.as_py())
 
 
+def arrow_cast(to: pa.DataType, arrow_data: pa.ChunkedArray) -> pa.ChunkedArray:
+    return pac.cast(arrow_data, to)
+
+
+def arrow_cast_integer(arrow_data: pa.ChunkedArray) -> pa.ChunkedArray:
+    arrow_int_type = ExpressionType.integer().to_arrow_type()
+    # FLOAT -> INTEGER cast is a truncation
+    if pa.types.is_floating(arrow_data.type):
+        truncated = pac.trunc(arrow_data)
+        replace_nan_with_0 = pac.if_else(pac.is_nan(truncated), 0, truncated).cast(arrow_int_type)
+        return replace_nan_with_0
+    return arrow_cast(arrow_int_type, arrow_data)
+
+
+def arrow_str_concat(prefix: pa.ChunkedArray, suffix: pa.ChunkedArray) -> pa.ChunkedArray:
+    return pac.binary_join_element_wise(prefix, suffix, pa.scalar(""))
+
+
 def _arr_unary_op(
     fn: Callable[..., pa.ChunkedArray],
 ) -> Callable[[DataBlock[ArrowArrType]], DataBlock[ArrowArrType]]:
@@ -777,6 +835,12 @@ class ArrowEvaluator(OperatorEvaluator["ArrowDataBlock"]):
     POSITIVE = ArrowDataBlock.identity
     ABS = _arr_unary_op(pac.abs)
     INVERT = _arr_unary_op(pac.invert)
+    CAST_INT = _arr_unary_op(arrow_cast_integer)
+    CAST_FLOAT = _arr_unary_op(partial(arrow_cast, ExpressionType.float().to_arrow_type()))
+    CAST_STRING = _arr_unary_op(partial(arrow_cast, ExpressionType.string().to_arrow_type()))
+    CAST_LOGICAL = _arr_unary_op(partial(arrow_cast, ExpressionType.logical().to_arrow_type()))
+    CAST_DATE = _arr_unary_op(partial(arrow_cast, ExpressionType.date().to_arrow_type()))
+    CAST_BYTES = _arr_unary_op(partial(arrow_cast, ExpressionType.bytes().to_arrow_type()))
     ADD = _arr_bin_op(pac.add)
     SUB = _arr_bin_op(pac.subtract)
     MUL = _arr_bin_op(pac.multiply)
@@ -792,6 +856,7 @@ class ArrowEvaluator(OperatorEvaluator["ArrowDataBlock"]):
     NEQ = _arr_bin_op(pac.not_equal)
     GT = _arr_bin_op(pac.greater)
     GE = _arr_bin_op(pac.greater_equal)
+    STR_CONCAT = _arr_bin_op(arrow_str_concat)
     STR_CONTAINS = _arr_bin_op(arrow_str_contains)
     STR_ENDSWITH = _arr_bin_op(arrow_str_endswith)
     STR_STARTSWITH = _arr_bin_op(arrow_str_startswith)
@@ -842,8 +907,8 @@ def assert_invalid_pylist_operation(*args, **kwargs):
     raise AssertionError(f"This is an invalid operation on a PyListDataBlock and should never be executing")
 
 
-def pylist_is_none(obj: Any):
-    return obj is None
+def pylist_is_none(input_block: PyListDataBlock) -> ArrowDataBlock:
+    return ArrowDataBlock(pa.chunked_array([[item is None for item in input_block.data]]))
 
 
 def pylist_if_else(
@@ -852,11 +917,53 @@ def pylist_if_else(
     return PyListDataBlock([xitem if c else yitem for c, xitem, yitem in zip_blocks_as_py(cond, x, y)])
 
 
+def _assert_and_identity(assert_type: type[T], obj: Any) -> T:
+    if not isinstance(obj, assert_type):
+        raise TypeError(f"Object {obj} is not of user-requested type {assert_type}")
+    return obj
+
+
+_PY_TO_PRIMITIVE_CAST_FUNCS: dict[ExpressionType, Callable] = {
+    ExpressionType.string(): str,
+    ExpressionType.integer(): int,
+    ExpressionType.float(): float,
+    ExpressionType.bytes(): bytes,
+    ExpressionType.logical(): bool,
+    ExpressionType.date(): partial(_assert_and_identity, datetime.date),
+}
+
+
+def pylist_cast(to: ExpressionType, pylist_block: PyListDataBlock) -> DataBlock:
+    if ExpressionType.is_primitive(to):
+        assert isinstance(to, PrimitiveExpressionType)
+        arrow_type = to.to_arrow_type()
+
+        assert to in _PY_TO_PRIMITIVE_CAST_FUNCS, f"Casting PY to {to} is not implemented"
+        cast_func = _PY_TO_PRIMITIVE_CAST_FUNCS[to]
+
+        return ArrowDataBlock(
+            pa.chunked_array([[cast_func(d) if d is not None else None for d in pylist_block.data]], type=arrow_type)
+        )
+
+    assert isinstance(to, PythonExpressionType)
+    for d in pylist_block.data:
+        if not (d is None or isinstance(d, to.python_cls)):
+            raise TypeError(f"Found object of type {type(d)} when casting types, expected {to}")
+
+    return pylist_block
+
+
 class PyListEvaluator(OperatorEvaluator["PyListDataBlock"]):
     NEGATE = make_map_unary(operator.neg)
     POSITIVE = PyListDataBlock.identity
     ABS = make_map_unary(operator.abs)
     INVERT = make_map_unary(operator.invert)
+    CAST_INT = partial(pylist_cast, ExpressionType.integer())
+    CAST_FLOAT = partial(pylist_cast, ExpressionType.float())
+    CAST_STRING = partial(pylist_cast, ExpressionType.string())
+    CAST_LOGICAL = partial(pylist_cast, ExpressionType.logical())
+    CAST_DATE = partial(pylist_cast, ExpressionType.date())
+    CAST_BYTES = partial(pylist_cast, ExpressionType.bytes())
     ADD = make_map_binary(operator.add)
     SUB = make_map_binary(operator.sub)
     MUL = make_map_binary(operator.mul)
@@ -873,7 +980,7 @@ class PyListEvaluator(OperatorEvaluator["PyListDataBlock"]):
     GT = make_map_binary(operator.gt)
     GE = make_map_binary(operator.ge)
 
-    IS_NULL = make_map_unary(pylist_is_none)
+    IS_NULL = pylist_is_none
     IF_ELSE = pylist_if_else
 
     # Placeholders: these change the cardinality of the block and should never be evaluated from the Evaluator
@@ -887,6 +994,7 @@ class PyListEvaluator(OperatorEvaluator["PyListDataBlock"]):
 
     # Unary operations that should never run on a PyListDataBlock because they are represented by
     # Arrow primitives and should always be housed in an ArrowDataBlock
+    STR_CONCAT = assert_invalid_pylist_operation
     STR_CONTAINS = assert_invalid_pylist_operation
     STR_ENDSWITH = assert_invalid_pylist_operation
     STR_STARTSWITH = assert_invalid_pylist_operation
