@@ -2,7 +2,7 @@ from __future__ import annotations
 
 import multiprocessing
 from dataclasses import dataclass
-from typing import Callable, ClassVar, Dict, List, Type
+from typing import Callable, ClassVar
 
 from daft.execution.execution_plan import ExecutionPlan
 from daft.execution.logical_op_runners import (
@@ -16,11 +16,18 @@ from daft.logical.logical_plan import LogicalPlan
 from daft.logical.optimizer import (
     DropRepartition,
     FoldProjections,
+    PruneColumns,
+    PushDownClausesIntoScan,
     PushDownLimit,
     PushDownPredicates,
 )
 from daft.resource_request import ResourceRequest
-from daft.runners.partitioning import PartID, PartitionManager, PartitionSet, vPartition
+from daft.runners.partitioning import (
+    PartID,
+    PartitionCacheEntry,
+    PartitionSet,
+    vPartition,
+)
 from daft.runners.profiler import profiler
 from daft.runners.runner import Runner
 from daft.runners.shuffle_ops import (
@@ -35,9 +42,9 @@ from daft.runners.shuffle_ops import (
 
 @dataclass
 class LocalPartitionSet(PartitionSet[vPartition]):
-    _partitions: Dict[PartID, vPartition]
+    _partitions: dict[PartID, vPartition]
 
-    def _get_all_vpartitions(self) -> List[vPartition]:
+    def _get_all_vpartitions(self) -> list[vPartition]:
         partition_ids = sorted(list(self._partitions.keys()))
         assert partition_ids[0] == 0
         assert partition_ids[-1] + 1 == len(partition_ids)
@@ -58,13 +65,16 @@ class LocalPartitionSet(PartitionSet[vPartition]):
     def __len__(self) -> int:
         return sum(self.len_of_partitions())
 
-    def len_of_partitions(self) -> List[int]:
+    def len_of_partitions(self) -> list[int]:
         partition_ids = sorted(list(self._partitions.keys()))
         return [len(self._partitions[pid]) for pid in partition_ids]
 
     def num_partitions(self) -> int:
         return len(self._partitions)
 
+    def wait(self) -> None:
+        pass
+
 
 class PyRunnerSimpleShuffler(Shuffler):
     def run(self, input: PartitionSet, num_target_partitions: int) -> PartitionSet:
@@ -105,8 +115,8 @@ class PyRunnerSortOp(PyRunnerSimpleShuffler, SortOp):
 class LocalLogicalPartitionOpRunner(LogicalPartitionOpRunner):
     def run_node_list(
         self,
-        inputs: Dict[int, PartitionSet],
-        nodes: List[LogicalPlan],
+        inputs: dict[int, PartitionSet],
+        nodes: list[LogicalPlan],
         num_partitions: int,
         resource_request: ResourceRequest,
     ) -> PartitionSet:
@@ -120,7 +130,7 @@ class LocalLogicalPartitionOpRunner(LogicalPartitionOpRunner):
 
 
 class LocalLogicalGlobalOpRunner(LogicalGlobalOpRunner):
-    shuffle_ops: ClassVar[Dict[Type[ShuffleOp], Type[Shuffler]]] = {
+    shuffle_ops: ClassVar[dict[type[ShuffleOp], type[Shuffler]]] = {
         RepartitionRandomOp: PyRunnerRepartitionRandom,
         RepartitionHashOp: PyRunnerRepartitionHash,
         CoalesceOp: PyRunnerCoalesceOp,
@@ -133,14 +143,14 @@ class LocalLogicalGlobalOpRunner(LogicalGlobalOpRunner):
         # NOTE: resource_request is ignored since there isn't any actual distribution of workloads in PyRunner
         return LocalPartitionSet({i: func(pset.get_partition(i)) for i in range(pset.num_partitions())})
 
-    def reduce_partitions(self, pset: PartitionSet, func: Callable[[List[vPartition]], ReduceType]) -> ReduceType:
+    def reduce_partitions(self, pset: PartitionSet, func: Callable[[list[vPartition]], ReduceType]) -> ReduceType:
         data = [pset.get_partition(i) for i in range(pset.num_partitions())]
         return func(data)
 
 
 class PyRunner(Runner):
     def __init__(self) -> None:
-        self._part_manager = PartitionManager(lambda: LocalPartitionSet({}))
+        super().__init__()
         self._part_op_runner = LocalLogicalPartitionOpRunner()
         self._global_op_runner = LocalLogicalGlobalOpRunner()
         self._optimizer = RuleRunner(
@@ -148,20 +158,29 @@ class PyRunner(Runner):
                 RuleBatch(
                     "SinglePassPushDowns",
                     Once,
-                    [PushDownPredicates(), FoldProjections(), DropRepartition()],
+                    [
+                        DropRepartition(),
+                        PushDownPredicates(),
+                        PruneColumns(),
+                        FoldProjections(),
+                        PushDownClausesIntoScan(),
+                    ],
                 ),
                 RuleBatch(
-                    "PushDownLimits",
+                    "PushDownLimitsAndRepartitions",
                     FixedPointPolicy(3),
-                    [PushDownLimit()],
+                    [PushDownLimit(), DropRepartition()],
                 ),
             ]
         )
 
-    def run(self, plan: LogicalPlan) -> PartitionSet:
-        plan = self._optimizer.optimize(plan)
-        # plan.to_dot_file()
-        exec_plan = ExecutionPlan.plan_from_logical(plan)
+    def optimize(self, plan: LogicalPlan) -> LogicalPlan:
+        return self._optimizer.optimize(plan)
+
+    def run(self, plan: LogicalPlan) -> PartitionCacheEntry:
+        optimized_plan = self.optimize(plan)
+
+        exec_plan = ExecutionPlan.plan_from_logical(optimized_plan)
         result_partition_set: PartitionSet
 
         # Check that the local machine has sufficient resources available for execution
@@ -175,15 +194,15 @@ class PyRunner(Runner):
                 raise RuntimeError(
                     f"Requested {resource_request.num_gpus} GPUs but found only {cuda_device_count()} available"
                 )
-
+        partition_intermediate_results: dict[int, PartitionSet] = {}
         with profiler("profile.json"):
             for exec_op in exec_plan.execution_ops:
 
                 data_deps = exec_op.data_deps
-                input_partition_set = {nid: self._part_manager.get_partition_set(nid) for nid in data_deps}
+                input_partition_set = {nid: partition_intermediate_results[nid] for nid in data_deps}
 
                 if exec_op.is_global_op:
-                    input_partition_set = {nid: self._part_manager.get_partition_set(nid) for nid in data_deps}
+                    input_partition_set = {nid: partition_intermediate_results[nid] for nid in data_deps}
                     result_partition_set = self._global_op_runner.run_node_list(
                         input_partition_set, exec_op.logical_ops
                     )
@@ -193,9 +212,11 @@ class PyRunner(Runner):
                     )
 
                 for child_id in data_deps:
-                    self._part_manager.rm(child_id)
+                    del partition_intermediate_results[child_id]
 
-                self._part_manager.put_partition_set(exec_op.logical_ops[-1].id(), result_partition_set)
+                partition_intermediate_results[exec_op.logical_ops[-1].id()] = result_partition_set
 
             last = exec_plan.execution_ops[-1].logical_ops[-1]
-            return self._part_manager.get_partition_set(last.id())
+            final_result = partition_intermediate_results[last.id()]
+            pset_id = self._part_set_cache.put_partition_set(final_result)
+            return pset_id
