@@ -1,8 +1,9 @@
 from __future__ import annotations
 
 from dataclasses import dataclass
-from typing import Any, Callable, ClassVar, Dict, List, Optional, Type, cast
+from typing import TYPE_CHECKING, Any, Callable, ClassVar, List, cast
 
+import pyarrow as pa
 import ray
 from loguru import logger
 
@@ -17,12 +18,21 @@ from daft.logical.logical_plan import LogicalPlan
 from daft.logical.optimizer import (
     DropRepartition,
     FoldProjections,
+    PruneColumns,
+    PushDownClausesIntoScan,
     PushDownLimit,
     PushDownPredicates,
 )
 from daft.resource_request import ResourceRequest
-from daft.runners.partitioning import PartID, PartitionManager, PartitionSet, vPartition
+from daft.runners.blocks import ArrowDataBlock, zip_blocks_as_py
+from daft.runners.partitioning import (
+    PartID,
+    PartitionCacheEntry,
+    PartitionSet,
+    vPartition,
+)
 from daft.runners.profiler import profiler
+from daft.runners.pyrunner import LocalPartitionSet
 from daft.runners.runner import Runner
 from daft.runners.shuffle_ops import (
     CoalesceOp,
@@ -33,17 +43,42 @@ from daft.runners.shuffle_ops import (
     SortOp,
 )
 
+if TYPE_CHECKING:
+    from ray.data.block import Block as RayDatasetBlock
+    from ray.data.dataset import Dataset as RayDataset
+
+
+@ray.remote
+def _make_ray_block_from_vpartition(partition: vPartition) -> RayDatasetBlock:
+    daft_blocks = {tile.column_name: tile.block for _, tile in partition.columns.items()}
+
+    all_arrow = all(isinstance(daft_block, ArrowDataBlock) for daft_block in daft_blocks.values())
+    if all_arrow:
+        return pa.Table.from_pydict({colname: daft_block.data for colname, daft_block in daft_blocks.items()})
+
+    colnames = list(daft_blocks.keys())
+    blocks = list(daft_blocks.values())
+    return [dict(zip(colnames, row_tuple)) for row_tuple in zip_blocks_as_py(*blocks)]
+
 
 @dataclass
 class RayPartitionSet(PartitionSet[ray.ObjectRef]):
-    _partitions: Dict[PartID, ray.ObjectRef]
+    _partitions: dict[PartID, ray.ObjectRef]
 
-    def _get_all_vpartitions(self) -> List[vPartition]:
+    def _get_all_vpartitions(self) -> list[vPartition]:
         partition_ids = sorted(list(self._partitions.keys()))
         assert partition_ids[0] == 0
         assert partition_ids[-1] + 1 == len(partition_ids)
         return cast(List[vPartition], ray.get([self._partitions[pid] for pid in partition_ids]))
 
+    def to_ray_dataset(self) -> RayDataset:
+        from ray.data import from_arrow_refs
+
+        blocks = [_make_ray_block_from_vpartition.remote(self._partitions[k]) for k in self._partitions.keys()]
+        # NOTE: although the Ray method is called `from_arrow_refs`, this method works also when the blocks are List[T] types
+        # instead of Arrow tables as the codepath for Dataset creation is the same.
+        return from_arrow_refs(blocks)
+
     def get_partition(self, idx: PartID) -> ray.ObjectRef:
         return self._partitions[idx]
 
@@ -59,19 +94,22 @@ class RayPartitionSet(PartitionSet[ray.ObjectRef]):
     def __len__(self) -> int:
         return sum(self.len_of_partitions())
 
-    def len_of_partitions(self) -> List[int]:
+    def len_of_partitions(self) -> list[int]:
         partition_ids = sorted(list(self._partitions.keys()))
 
         @ray.remote
         def remote_len(p: vPartition) -> int:
             return len(p)
 
-        result: List[int] = ray.get([remote_len.remote(self._partitions[pid]) for pid in partition_ids])
+        result: list[int] = ray.get([remote_len.remote(self._partitions[pid]) for pid in partition_ids])
         return result
 
     def num_partitions(self) -> int:
         return len(self._partitions)
 
+    def wait(self) -> None:
+        ray.wait([o for o in self._partitions.values()])
+
 
 class RayRunnerSimpleShuffler(Shuffler):
     def run(self, input: PartitionSet, num_target_partitions: int) -> PartitionSet:
@@ -88,7 +126,7 @@ class RayRunnerSimpleShuffler(Shuffler):
         @ray.remote(num_returns=num_target_partitions)
         def map_wrapper(input: vPartition):
             output_dict = self.map_fn(input=input, output_partitions=num_target_partitions, **map_args)
-            output_list: List[Optional[vPartition]] = [None for _ in range(num_target_partitions)]
+            output_list: list[vPartition | None] = [None for _ in range(num_target_partitions)]
             for part_id, part in output_dict.items():
                 output_list[part_id] = part
 
@@ -141,15 +179,15 @@ class RayRunnerSortOp(RayRunnerSimpleShuffler, SortOp):
 def _ray_partition_single_part_runner(
     *input_parts: vPartition,
     op_runner: RayLogicalPartitionOpRunner,
-    input_node_ids: List[int],
-    nodes: List[LogicalPlan],
+    input_node_ids: list[int],
+    nodes: list[LogicalPlan],
     partition_id: int,
 ) -> vPartition:
     input_partitions = {id: val for id, val in zip(input_node_ids, input_parts)}
     return op_runner.run_node_list_single_partition(input_partitions, nodes=nodes, partition_id=partition_id)
 
 
-def _get_ray_task_options(resource_request: ResourceRequest) -> Dict[str, Any]:
+def _get_ray_task_options(resource_request: ResourceRequest) -> dict[str, Any]:
     options = {}
     if resource_request.num_cpus is not None:
         options["num_cpus"] = resource_request.num_cpus
@@ -161,8 +199,8 @@ def _get_ray_task_options(resource_request: ResourceRequest) -> Dict[str, Any]:
 class RayLogicalPartitionOpRunner(LogicalPartitionOpRunner):
     def run_node_list(
         self,
-        inputs: Dict[int, PartitionSet],
-        nodes: List[LogicalPlan],
+        inputs: dict[int, PartitionSet],
+        nodes: list[LogicalPlan],
         num_partitions: int,
         resource_request: ResourceRequest,
     ) -> PartitionSet:
@@ -179,7 +217,7 @@ class RayLogicalPartitionOpRunner(LogicalPartitionOpRunner):
 
 
 class RayLogicalGlobalOpRunner(LogicalGlobalOpRunner):
-    shuffle_ops: ClassVar[Dict[Type[ShuffleOp], Type[Shuffler]]] = {
+    shuffle_ops: ClassVar[dict[type[ShuffleOp], type[Shuffler]]] = {
         RepartitionRandomOp: RayRunnerRepartitionRandom,
         RepartitionHashOp: RayRunnerRepartitionHash,
         CoalesceOp: RayRunnerCoalesceOp,
@@ -192,23 +230,19 @@ class RayLogicalGlobalOpRunner(LogicalGlobalOpRunner):
         remote_func = ray.remote(func).options(**_get_ray_task_options(resource_request))
         return RayPartitionSet({i: remote_func.remote(pset.get_partition(i)) for i in range(pset.num_partitions())})
 
-    def reduce_partitions(self, pset: PartitionSet, func: Callable[[List[vPartition]], ReduceType]) -> ReduceType:
-        # @ray.remote
-        # def remote_func(*parts: vPartition) -> ReduceType:
-        #     return func(list(parts))
-
+    def reduce_partitions(self, pset: PartitionSet, func: Callable[[list[vPartition]], ReduceType]) -> ReduceType:
         data = [pset.get_partition(i) for i in range(pset.num_partitions())]
         result: ReduceType = func(ray.get(data))
         return result
 
 
 class RayRunner(Runner):
-    def __init__(self, address: Optional[str]) -> None:
+    def __init__(self, address: str | None) -> None:
+        super().__init__()
         if ray.is_initialized():
             logger.warning(f"Ray has already been initialized, Daft will reuse the existing Ray context")
         else:
             ray.init(address=address)
-        self._part_manager = PartitionManager(lambda: RayPartitionSet({}))
         self._part_op_runner = RayLogicalPartitionOpRunner()
         self._global_op_runner = RayLogicalGlobalOpRunner()
         self._optimizer = RuleRunner(
@@ -216,27 +250,44 @@ class RayRunner(Runner):
                 RuleBatch(
                     "SinglePassPushDowns",
                     Once,
-                    [PushDownPredicates(), FoldProjections(), DropRepartition()],
+                    [
+                        DropRepartition(),
+                        PushDownPredicates(),
+                        PruneColumns(),
+                        FoldProjections(),
+                        PushDownClausesIntoScan(),
+                    ],
                 ),
                 RuleBatch(
-                    "PushDownLimits",
+                    "PushDownLimitsAndRepartitions",
                     FixedPointPolicy(3),
-                    [PushDownLimit()],
+                    [PushDownLimit(), DropRepartition()],
                 ),
             ]
         )
 
-    def run(self, plan: LogicalPlan) -> PartitionSet:
-        plan = self._optimizer.optimize(plan)
+    def put_partition_set_into_cache(self, pset: PartitionSet) -> PartitionCacheEntry:
+        if isinstance(pset, LocalPartitionSet):
+            pset = RayPartitionSet({pid: ray.put(val) for pid, val in pset._partitions.items()})
+
+        return self._part_set_cache.put_partition_set(pset=pset)
+
+    def optimize(self, plan: LogicalPlan) -> LogicalPlan:
+        return self._optimizer.optimize(plan)
+
+    def run(self, plan: LogicalPlan) -> PartitionCacheEntry:
+        plan = self.optimize(plan)
         exec_plan = ExecutionPlan.plan_from_logical(plan)
         result_partition_set: PartitionSet
+        partition_intermediate_results: dict[int, PartitionSet] = {}
         with profiler("profile.json"):
             for exec_op in exec_plan.execution_ops:
+
                 data_deps = exec_op.data_deps
-                input_partition_set = {nid: self._part_manager.get_partition_set(nid) for nid in data_deps}
+                input_partition_set = {nid: partition_intermediate_results[nid] for nid in data_deps}
 
                 if exec_op.is_global_op:
-                    input_partition_set = {nid: self._part_manager.get_partition_set(nid) for nid in data_deps}
+                    input_partition_set = {nid: partition_intermediate_results[nid] for nid in data_deps}
                     result_partition_set = self._global_op_runner.run_node_list(
                         input_partition_set, exec_op.logical_ops
                     )
@@ -244,12 +295,13 @@ class RayRunner(Runner):
                     result_partition_set = self._part_op_runner.run_node_list(
                         input_partition_set, exec_op.logical_ops, exec_op.num_partitions, exec_op.resource_request()
                     )
-                del input_partition_set
+
                 for child_id in data_deps:
-                    self._part_manager.rm(child_id)
-                self._part_manager.put_partition_set(exec_op.logical_ops[-1].id(), result_partition_set)
-                del result_partition_set
-            last_id = exec_plan.execution_ops[-1].logical_ops[-1].id()
-            last_pset = self._part_manager.get_partition_set(last_id)
-            self._part_manager.clear()
-            return last_pset
+                    del partition_intermediate_results[child_id]
+
+                partition_intermediate_results[exec_op.logical_ops[-1].id()] = result_partition_set
+
+            last = exec_plan.execution_ops[-1].logical_ops[-1]
+            final_result = partition_intermediate_results[last.id()]
+            pset_id = self._part_set_cache.put_partition_set(final_result)
+            return pset_id
