@@ -3,7 +3,6 @@ from __future__ import annotations
 import datetime
 from dataclasses import dataclass
 from enum import Enum
-from typing import Dict, Type
 
 import pyarrow as pa
 
@@ -54,7 +53,7 @@ class ExpressionType:
         return _TYPE_REGISTRY["null"]
 
     @staticmethod
-    def from_py_type(obj_type: Type) -> ExpressionType:
+    def from_py_type(obj_type: type) -> ExpressionType:
         """Gets the appropriate ExpressionType from a Python object, or _TYPE_REGISTRY["unknown"]
         if unable to find the appropriate type. ExpressionTypes.Python is never returned.
         """
@@ -72,6 +71,10 @@ class ExpressionType:
             return ExpressionType.python_object()
         return _PYARROW_TYPE_TO_EXPRESSION_TYPE[datatype]
 
+    def to_arrow_type(self) -> pa.DataType:
+        assert not ExpressionType.is_py(self), f"Cannot convert {self} to an Arrow type"
+        return _EXPRESSION_TYPE_TO_PYARROW_TYPE[self]
+
 
 @dataclass(frozen=True, eq=True)
 class PrimitiveExpressionType(ExpressionType):
@@ -91,19 +94,16 @@ class PrimitiveExpressionType(ExpressionType):
     def __repr__(self) -> str:
         return self.enum.name
 
-    def to_arrow_type(self) -> pa.DataType:
-        return _EXPRESSION_TYPE_TO_PYARROW_TYPE[self]
-
 
 @dataclass(frozen=True, eq=True)
 class PythonExpressionType(ExpressionType):
-    python_cls: Type
+    python_cls: type
 
     def __repr__(self) -> str:
         return f"PY[{self.python_cls.__name__}]"
 
 
-_TYPE_REGISTRY: Dict[str, ExpressionType] = {
+_TYPE_REGISTRY: dict[str, ExpressionType] = {
     "unknown": PrimitiveExpressionType(PrimitiveExpressionType.TypeEnum.UNKNOWN),
     "integer": PrimitiveExpressionType(PrimitiveExpressionType.TypeEnum.INTEGER),
     "float": PrimitiveExpressionType(PrimitiveExpressionType.TypeEnum.FLOAT),
